/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*****************************************************************************
  program name: gl_auto.c  (running as a version of lm_auto).
******************************************************************************/

/**
 * @defgroup lm_auto lm_auto
 * @ingroup autozyg
 * @{
 */

#define PNAM  "lm_auto"

#include "autozyg.h"

void  gl_do_auto_sweep (MSD_PTR, int);
void  prnt_scorF (int);

static FILE *oScorF = NULL;

static int scored_iterations;           /* Number of scored iterations.       */

static int mc_itno;
static int nsw;

                                        /* Counts and log-probabilities.      */
static double *ctot;                    /* CDLL by component.                 */
static double **prob;

static Indlist *pind;                   /* Pointer to component indlist.      */

int main (int argc, char *argv[])
{
   FILE *oSeedF;

   MSD_PTR p_msd;

   int   output_iterations;             /* Total number of iterations in      */
   /*                                      which gl_auto is to add output to  */
   /*                                      the scores file.                   */

   int   nCompnt1;                      /* Number of components to be sampled */
   /*                                      from the exact distribution.       */
   int   comp_no;                       /* Current component number.          */
   int   iteration;

   /* ................................  SET-UP   ............................ */

   nsw = setupAutozyg (argc, argv, PNAM, 0);

   /* It doesn't make a whole lot of sense to continue running gl_auto unless */
   /* at least one of the two possible types of output has been selected by   */
   /* the gl_auto user.                                                       */

   if (PrintFGLs == N && PrintMIs == N)
      give_up (ERR_nmi);

   if (DebugMain)
      list_switches (nsw);

   list_nghds ();
   printf ("\n");
   fflush (stdout);

   setupSampling ();

   nCompnt1 = 0;
   if (!GlobalMcmc)
   {
      /* Count the number of components for which sampling from the exact     */
      /* distribution is to be done.                                          */
      for (comp_no = 1; comp_no <= nCompnt; ++comp_no)
         if (cmplist[comp_no].nmeioses <= MaxNMexact)
            ++nCompnt1;
   }

   prob = setupInherit (McmcMarkersOnly);

   if (MMSampler == Y)
   {
      /* Set up for the multiple-meiosis sampler.                             */
      Max_update = 8;
      P_meio = 0.2;
      P_ind = 0.3;
      P_sib = 0.3;
      P_3gen = 0.2;

      setup_marriage ();
      prnt_marriage ();
   }

   else if (MMSampler == N)
   {
      /* Set up for the old M-sampler.                                        */
      Max_update = 1;
      P_meio = 1.0;
      P_ind = 0.0;
      P_sib = 0.0;
      P_3gen = 0.0;
   }

   /* Open output scores file, if requested to do so.                         */

   if (FilIt && !oScorF)
   {
      if (!(oScorF = open_fil (OSCORE)))
         give_up (ERR_qit);
      if (GlobalMcmc)
         printf ("\n");
      fflush (stdout);
   }

   /* Calculate the total number of iterations in which gl_auto is to add     */
   /* output to the scores file.                                              */

   output_iterations = NIt / (ScorIt * max (FilIt, 1));

   if (PrintLevel >= 4)
   {
      ctot = malloc (n_compnts * sizeof (double));
      alloc_chk (WHERE, ctot);

      p_msd = alloc_msd ();
      alloc_logprob ();
   }

   if (!GlobalMcmc)
   {
      for (comp_no = 1; comp_no <= n_compnts; ++comp_no)
      {
         if (PrintLevel >= 4 && comp_no > 1)
         {
            /* Initialize variables which need to be zeroed for each new      */
            /* component being sampled.  Note that p_msd->temp_log_prob_seg   */
            /* does not need to be zeroed.                                    */

            zero_mcmc_stats (p_msd);
         }

         scored_iterations = 0;         /*Initialize scored iterations count. */

         printf ("\n ====== Component %d will be determined using", comp_no);

         if (cmplist[comp_no].nmeioses <= MaxNMexact)
         {
            printf (" IID exact sampling ======\n\n\n");
            fflush (stdout);

            /* Forward computation.                                           */
            exact_smpl (1, 0, comp_no, 0, NMarks, nFGL, prob, NMarks, ctot);

            /* Perform IID sampling backwards.                                */
            for (iteration = 1; iteration <= output_iterations; ++iteration)
            {
               exact_smpl (0, 1, comp_no, 0, NMarks, nFGL, prob, NMarks, ctot);

               if (PrintLevel >= 4)
                  update_counts ("C", p_msd, comp_no, 1, prob, ctot);

               ++scored_iterations;

               prnt_scorF (comp_no);
            }
         }

         else
         {
            /* cmplist[comp_no].nmeioses > MaxNMexact, so sampling is to be   */
            /* done separately for this component.                            */

            printf (" MCMC ======\n\n\n");

            prntStatusBMC ();           /* Print status for beginning MCMC.   */
            fflush (stdout);

            /* Setup variables for LM-multiple sampling: LPT Apr 2006.        */
            setup_sampling_cmp (comp_no, MMSampler, NMarks);

            /* ................  BURN-IN  .................................   */

            /* Now do burn-in iterations.                                     */

            printf (" Now starting %d iterations of burn-in\n", NBrnIt);
            for (iteration = 1; iteration <= NBrnIt; ++iteration)
               sampling_cmp (iteration, comp_no, McmcMarkersOnly, MMSampler,
                             NMarks, nFGL, prob);
            printf ("\n");
            fflush (stdout);

            /* Now do the regular MCMC sampling, and score the results.       */

            printf (" Now starting %d iterations of MCMC\n", NIt);
            fflush (stdout);

            for (mc_itno = 1; mc_itno <= NIt; ++mc_itno)
               gl_do_auto_sweep (p_msd, comp_no);

            /* Print status that we have finished with the MCMC.              */
            printf ("\n\n");
            prntStatusFMC ();

            free_sampling_cmp ();
         }

         /* Now print out the sampling statistics for the current component.  */

         pr_cmpnthead (comp_no - 1);
         pr_comp_mcmc_stats (p_msd, 1, scored_iterations, comp_no, 0);
         fflush (stdout);
      }
   }

   else if (GlobalMcmc)
   {
      prntStatusBMC ();                 /* Print status for beginning MCMC.   */
      fflush (stdout);

      /* Setup variables for LM-multiple sampling: LPT Apr 2006.              */
      setup_sampling_cmp (0, MMSampler, NMarks);

      /* ................  BURN-IN  .......................................   */

      /* Now do burn-in iterations.                                           */

      printf (" Now starting %d iterations of burn-in\n", NBrnIt);
      for (iteration = 1; iteration <= NBrnIt; ++iteration)
         sampling_cmp (iteration, 0, McmcMarkersOnly, MMSampler, NMarks, nFGL,
                       prob);
      printf ("\n");
      fflush (stdout);

      scored_iterations = 0;

      /* ................  COMPUTE USING MCMC  ............................   */

      /* Now do the regular MCMC sampling, and score the results.             */

      printf (" Now starting %d iterations of MCMC\n", NIt);
      fflush (stdout);

      for (mc_itno = 1; mc_itno <= NIt; ++mc_itno)
         gl_do_auto_sweep (p_msd, 0);

      /* Print status that we have finished with the MCMC.                    */
      printf ("\n\n");
      prntStatusFMC ();
      fflush (stdout);

      free_sampling_cmp ();

      /* Now print out the sampling statistics for each component.            */

      for (comp_no = 1; comp_no <= n_compnts; ++comp_no)
      {
         pr_cmpnthead (comp_no - 1);
         pr_comp_mcmc_stats (p_msd, 1, scored_iterations, comp_no, comp_no - 1);
         fflush (stdout);
      }
   }

   /* Free up allocated memory.                                               */
   free_xxgg ();

   if (PrintLevel >= 4)
   {
      free_1d ((void *) &ctot);
      free_msd (p_msd);
      free_logprob ();
   }

   locus_peel_free ();
   free_allocate_loci ();
   free_chromo (0);
   free_gen_pen ();
   free_seg_pen (NMarks + 1, &prob);
   free_msmpl ();
   free_traits ();

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }
   if (FilIt)
      fclose (oScorF);

   if (Scan)
      free_perm ();

   exit (EXIT_SUCCESS);
}

void gl_do_auto_sweep (MSD_PTR p_msd, int cmp)
{
   int   i_ndx;
   int   i1;

   /* Now do the next MCMC sampling iteration.                                */

   sampling_cmp (mc_itno, cmp, McmcMarkersOnly, MMSampler, NMarks, nFGL, prob);

   if (DebugMain)
   {
      fprintf (stderr, "\nAt iteration %d,\n", mc_itno);
      dbgXXGG ();
   }

   if ((mc_itno % ScorIt) == 0)
   {
      if (PrintLevel >= 4)
      {
         update_counts ("C", p_msd, cmp, 1, prob, ctot);
         ++scored_iterations;
      }

      /* Test for output to scores file.                                      */
      if (FilIt && ((mc_itno % (ScorIt * FilIt)) == 0))
      {
         if (cmp > 0)
            prnt_scorF (cmp);

         else if (cmp == 0)
         {
            if (PrintMIs == Y)
            {
               for (i1 = 1; i1 <= nsw; ++i1)
               {
                  i_ndx = (meiosis[i1].ind).index;
                  if (meiosis[i1].sex == 0)
                     prnt_xxx (oScorF, xx0, i_ndx);
                  else
                     prnt_xxx (oScorF, xx1, i_ndx);
               }
            }
            else if (PrintFGLs == Y)
            {
               for (i_ndx = 0; i_ndx < pedsize; ++i_ndx)
               {
                  prnt_fgl (oScorF, gg0, i_ndx);
                  prnt_fgl (oScorF, gg1, i_ndx);
               }
            }
         }

      }
   }

   return;
}

/* This subroutine was added by SML in March, 2011.                           */

void prnt_scorF (int cmp)
{
   int   last_ndx;
   int   first_meiosis;
   int   last_meiosis;
   int   i_ndx;
   int   i1;

   if (PrintMIs == Y)
   {
      last_ndx = cmplist[cmp].nmeioses - 1;
      first_meiosis = cmplist[cmp].meioses[0];
      last_meiosis = cmplist[cmp].meioses[last_ndx];
      for (i1 = first_meiosis; i1 <= last_meiosis; ++i1)
      {
         i_ndx = (meiosis[i1].ind).index;
         if (meiosis[i1].sex == 0)
            prnt_xxx (oScorF, xx0, i_ndx);
         else
            prnt_xxx (oScorF, xx1, i_ndx);
      }
   }

   else if (PrintFGLs == Y)
   {
      pind = cmplist[cmp].people->next;
      while (pind)
      {
         i_ndx = pind->current->index;
         {
            prnt_fgl (oScorF, gg0, i_ndx);
            prnt_fgl (oScorF, gg1, i_ndx);
         }
         pind = pind->next;
      }
   }

   return;
}
