/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*
 *       msample.h:  header file for the Msample/ library                     *
 *                -  included by Msample/ (meiosis sampling) functions        *
 *                   and by main programs calling Msample/ functions          *
 *         Msample library combined with Lsample library: EAT June 2002       *
 *            many Msample routines moved to new Markers library              *
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

#ifndef MSAMPLE_H
#define MSAMPLE_H

/**
 * @addtogroup sample
 * @{
 */

#include "incs.h"
#include "nghds.h"
#include "markers.h"
#include "rans.h"
#include "mulmsample.h"

struct gp_rfunc {
        struct gp_rfunc *next;
        int *inv;
        double *p;
        int n_inv,nc,flag;
};

struct gp_rfunc_ptr {
        struct gp_rfunc_ptr *next;
        struct gp_rfunc *rf;
};

struct gp_rfnode {
        struct gp_rfnode *next;
        struct gp_rfunc *rf;
        int y;
};

struct deg_list {
        struct deg_list *next,*prev,*abs_list;
        int deg,gene;
};


/* Maximum number of genes in an RFunction for gen_pen()
 * Note that memory and patience will run out long before this does...       */
#define MAX_GP_RF_GENES 64

/* MaxNMexact replaced constant MAXNMexact in July of 2008 by EAT.  In order  */
/* to allow the parser to refer to MaxNMexact, it should be declared as a     */
/* GLOBAL variable in globs.h.  This was done by SML in November of 2010.     */

void  gene_update (int, int, int);

/* fwd-back-smple routines: multiple meiosis versions replace old ones */
/* Transferred from MulMsample library  EAT Aug 2006                   */
/* added routine fwd_bk combine() for P(S|Y);  EAT Dec 2008            */
double fwd_compute (int, int, int, double *, int *, double *,
                     double *, double *);
void   back_sample (int, int, int, int *, double *, int *,
                     double *, int *);
void fwd_bk_combine (int, int, int, int,  double *,
                  int *, double *, double *);

/* Routines in msmpl_subs.c, moved from MulMsample, EAT 2006  */
/*   Put into separate routines using LPT code:  EAT June 2006 */
void   biadd (int *, int m, int n);
long int pow2 (int);
void store_pen (int, int, int , int *, int *, int, int,  double **);
double calc_fwd (int, int, int, int *, double **, double **);
void bck_smpl (int, int, int, int *, double **, int **);
void assign_reset (int, int, int, int, int *, int *, int, double **, int **);
/*  Exact_lods and iid-sampling: transferred to Sample:  EAT Aug 2006 */
/*  file: exact_lods.c; exact_lods()  LPT March 2006 */
/*  and associated alloc/free routines,          added EAT May 2006 */
/*  and iid-sampling routines                    added EAT May 2006 */
/*  exact_lods modified to take loc_left and rec_left info : EAT Sept 2006 */
/*  exact_lods modified to do marker computes once only: EAT April 2007 */
/*  exact_ivprob added; EAT Dec 2008 */
void exact_smpl (int, int, int, int, int, int, double **, int, double *);
void iid_sample (int, int, int, int, double ** );
/*  The exact_lods declaration was modified and then transferred to the       */
/*  lsample.h header file,                       by SML, in July of 2010.     */
void exact_ivprob (int, int, int, int, double **);
void alloc_exact (int);
void free_exact (int);
void set_unif (double *, int, double);      /* added EAT: April 2007     */

/*  end of transferred routines */

void  set_FGL (int);
void  reset_FGL (int, int);               /* added EAT 12/2008   */
void  realize_null (int, int);
void  realise (int, double ***, int);
void  set_trait_segs (int);
void  next_marker_segs (int, int, int, int);
void  set_genotypes (int, int, double *);
void  free_allocate_loci (void);
void  do_msmpl (int, int, int, int, int, double **, int);
void  alloc_msmpl (void);
void  free_msmpl (void);

void  full_seg_prob (int, int, int ***, double **);
double diff_seg_prob (int, int, int, int *, double **);
double lik_rat (int, int, int, int);
double full_pen_seg_lod (int, int, int, int, int ***);

/* new from LPT, July 2007 */
void full_seg_prob_cmp (int, int, int **, double *);

/* following appear to be obsolete: EAT June 2002   */
/* void  print_segs (int);
 * double new_pen (int, int *[2], int ***, double *, int, int, int *);
 * double prob_data (int, int, int, int);
 */

/* New from Simon  06/20/2001.                                             */
/*  all seg_pen/gen_pen routines moved to lsample.h;    EAT  4/2010        */

/* new sampling routine;  EAT Sep 2002. Updated to remve y, EAT Sep 2004 */

void sampling (int, int, int, int, int, double **);

void setMSampling (void);

/** @} */

#endif
