/***************************************************************************/
/*  Program to do a basic simulation and analysis of IBD inference with LD */
/*  Routines to process input data and make haplotypes                     */
/*  Contains:  input_haps ():  reads in the initial haplotypes             */
/*             make_haplo ():  recycles and subselects SNPs for analysis   */
/*           make_haplo_segs():  copies segments from constr ibd to haplos */
/*           make_data():  set the haplo data from constructed haplos      */
/*           select_snps(): iselects SNPs for analysis from input block    */
/*           print_haplo(): prints the snp data on a haplotype             */
/***************************************************************************/

void make_pair_data (int *hd, int **hm, int nd, int skp, int stmk, int *prpr)
/* set data for pair of haplotypes (prpr) only  */
{
   int   i, j, k1, k2;
   k1 = prpr[0];
   k2 = prpr[1];
   for (i = 0; i < nd; ++i)
   {
      j = i * skp + stmk;               /* Setting up marker j.               */
      hd[i] = -1;
      if ((hm[k1][j] == 1) && (hm[k2][j] == 1))
         hd[i] = 0;
      if ((hm[k1][j] == 1) && (hm[k2][j] == 2))
         hd[i] = 1;
      if ((hm[k1][j] == 2) && (hm[k2][j] == 1))
         hd[i] = 2;
      if ((hm[k1][j] == 2) && (hm[k2][j] == 2))
         hd[i] = 3;
   }
   return;
}

void make_data (int *hd, int **hm, int nh, int kk, int nd, int skp, int stmk)
/* Make data length nd, from kk'th quartet of haplotypes                      */
/* using every skp'th marker from nh haplotypes hm.                           */
/*  Allow for missing data --- hd =-1 if any hm=0.                            */
/*  PhaseofData is (now Global) genotypic data indicator for analysis.        */
{
   int   i, j, k;
   for (i = 0; i < nd; ++i)
   {
      j = i * skp + stmk;               /* Setting up marker j.               */
      hd[i] = 0;
      for (k = kk * nh; k < nh + kk * nh; ++k)
      {
         if (hm[k][j] == 0)
         {
            hd[i] = -1;
            break;
         }
         hd[i] = 2 * hd[i] + (hm[k][j] - 1);
      }
   }

   return;
}

void print_haplo (FILE * fptr, int *hap, int mrk)
{
   int   j;
   fprintf (fptr, "\n");
   for (j = 0; j < mrk; ++j)
      fprintf (fptr, "%2d", hap[j]);

   return;
}
