/*  routines included in gl_lods programs, mainly for processing
 * DGL file  etc.
 */

int dgl_index ()
/* First we will go through one cycle of DGL graph, and find index in nghd.
   If not in nghd we will drop; if in nghd but with 0 index, we will drop.
   We will set up to indicate in nghd which are found;
     then go through nghds to check all obsved have been found--else WARN*/
/* Mar 2011: we will first count the number found in the gl_auto file     */
{
   int   i, id, j, sw, err, nn;

   int  *nghd_index;
   char  nam1[MAXNAMLEN + 1], nam2[MAXNAMLEN + 1];
   FILE *fp;

   if (!(fp = open_fil (XTRA)))         /* This is raw DGL graphs  */
      give_up (ERR_qit);
   nn = 0;
   i = 0;
   read_name2 (fp, nam1, &err);         /* first line of DGL file  */
   fscanf (fp, " %*d %*d %d", &sw);
   for (j = 0; j < sw; ++j)
      fscanf (fp, " %*d %*d");
   while (i == 0)
   {
      ++nn;
      read_name2 (fp, nam2, &err);
      fscanf (fp, " %*d %*d %d", &sw);
      for (j = 0; j < sw; ++j)
         fscanf (fp, " %*d %*d");
      if (read_name2 (fp, nam2, &err) == EOF)
         break;
      fscanf (fp, " %*d %*d %d", &sw);
      for (j = 0; j < sw; ++j)
         fscanf (fp, " %*d %*d");
      if (!strncmp (nam1, nam2, MAXNAMLEN + 1))
         break;
   }
   printf (" Number of individuals in dgl-graph file is %d\n", nn);
   fclose (fp);                         /* Individuals counted.               */

   DGL_names = malloc (nn * sizeof (char *));
   alloc_chk (WHERE, DGL_names);
   for (i = 0; i < nn; ++i)
   {
      DGL_names[i] = malloc ((MAXNAMLEN + 1) * sizeof (char));
      alloc_chk (WHERE, DGL_names[i]);
   }
   DGL_index = malloc (nn * sizeof (int));

   nghd_index = malloc (pedsize * sizeof (int));
   for (i = 0; i < pedsize; ++i)
      nghd_index[i] = 0;

   if (!(fp = open_fil (XTRA)))         /* This is raw DGL graphs  */
      give_up (ERR_qit);

   for (i = 0; i < nn; ++i)             /* going through DGLgraph */
   {
      read_name2 (fp, nam1, &err);
      fscanf (fp, " %*d %*d %d", &sw);
      for (j = 0; j < sw; ++j)
         fscanf (fp, " %*d %*d");
      read_name2 (fp, nam2, &err);
      fscanf (fp, " %*d %*d %d", &sw);
      for (j = 0; j < sw; ++j)
         fscanf (fp, " %*d %*d");
      fflush (stdout);
      if (!strncmp (nam1, nam2, MAXNAMLEN + 1))
         strncpy (DGL_names[i], nam1, MAXNAMLEN + 1);
      else
      {
         printf ("\n Paired names %s and %s in DGL file do not match",
                 nam1, nam2);
         exit (97);
      }
      id = get_posit (nam1);
      if (id == INVALID_INDEX)
         DGL_index[i] = INVALID_INDEX;
      else
      {
         nghd_index[id] = 1;
         if (nghd[id].obs)
            DGL_index[i] = id;
         else
            DGL_index[i] = INVALID_INDEX;
      }
   }

   if (read_name2 (fp, nam1, &err) != EOF)
   {
      if (strncmp (nam1, DGL_names[0], MAXNAMLEN + 1))
      {
         printf ("\n Next name %s does not match first name %s in DGL file",
                 nam1, DGL_names[0]);
         exit (97);
      }
   }

   for (i = 0; i < pedsize; ++i)        /* Now check nghd_index[].            */
      if ((nghd_index[i] == 0) && (nghd[i].obs == 1))
      {
         nghd[i].obs = 0;
         printf ("\n Observed individual %s is not in DGL file:", nghd[i].self);
         printf ("\n   The trait data on %s will be ignored (W)\n",
                 nghd[i].self);
      }
   printf ("\n");

   free (nghd_index);
   fclose (fp);

   return (nn);
}

int copy_line (FILE * fp_in, FILE * fp_out, int tt, int ii)
{
   int   j, id, k, sw, item, cc, err;
   char  nam1[MAXNAMLEN + 1];

   cc = 0;
   id = DGL_index[ii];
   if (id == INVALID_INDEX)
      for (j = 0; j < 2; ++j)
      {
         if (read_name2 (fp_in, nam1, &err) == EOF)
         {
            cc = -1;
            return cc;
         }
         fscanf (fp_in, " %*d %*d %d", &sw);
         for (k = 0; k < sw; ++k)
            fscanf (fp_in, " %*d %*d");
      }
   else
   {
      for (j = 0; j < 2; ++j)
      {
         if (read_name2 (fp_in, nam1, &err) == EOF)
         {
            cc = -1;
            return cc;
         }
         if (strncmp (nam1, nghd[id].self, MAXNAMLEN + 1))
         {
            printf
               ("\n Problem: name %s does not match %s at DGL posit %d in graph %d",
                nam1, nghd[id].self, ii, tt);
            exit (97);
         }
         fprintf (fp_out, " %d", id);
         for (k = 0; k < 2; ++k)
         {
            fscanf (fp_in, " %d", &item);
            fprintf (fp_out, " %d", item);
         }
         if (item > cc)
            cc = item;

         fscanf (fp_in, " %d", &sw);
         fprintf (fp_out, " %d", sw);
         for (k = 0; k < sw; ++k)
         {
            fscanf (fp_in, " %d", &item);
            fprintf (fp_out, " %d", item); /* switch point  */
            fscanf (fp_in, " %d", &item);
            fprintf (fp_out, " %d", item); /* switch FGL    */
            if (item > cc)
               cc = item;
         }
         fprintf (fp_out, "\n");
      }
   }

   return cc;
}

/* Go through the DGL graph file again, stripping unobserved individuals and  */
/* outputting by pedigree index.  Also track largest FGL for nFGL.            */
/* nn is the number of individuals in each DGL file (added March 2011, EAT).  */

int dgl_strip (int *n_dgl, int nn)
{
   int   i, t, cc, ctot, max_fgl;
   FILE *fp, *oScorF;

   ctot = 0;
   max_fgl = 0;

   if (!(fp = open_fil (XTRA)))         /* The input is the raw DGL graphs.   */
      give_up (ERR_qit);

   if (!(oScorF = open_fil (OSCORE)))   /* This is processed DGL graphs.      */
      give_up (ERR_qit);

   for (t = 0; t < NIt; ++t)            /* There are NIt DGLgraph's in  file. */
   {
      for (i = 0; i < nn; ++i)          /* going through DGLgraph */
      {
         cc = copy_line (fp, oScorF, t, i); /* return max fgl no */
         if (cc < 0)
            break;
         if (cc > max_fgl)
            max_fgl = cc;
         if (cc > 0)
            ++ctot;
      }
      if (cc < 0)
         break;
   }

   *n_dgl = t;                          /* number dgl grphs read */

   ctot = ctot / (*n_dgl);              /* number individuals per new grph */
   nFGL = max_fgl;                      /* reset nFGL   */

   fclose (fp);
   fclose (oScorF);

   return ctot;
}

void set_line (FILE * fp, int **ggg)
{
   int   j, k, id, fgl, fgl2, sw, swp;

   fscanf (fp, " %d %*d %d %d", &id, &fgl, &sw);
   if (!sw)
      for (j = 0; j < NMarks; ++j)
         ggg[j][id] = fgl;

   else
   {
      j = 0;
      for (k = 0; k < sw; ++k)
      {
         fscanf (fp, "%d %d", &swp, &fgl2);
         while ((j < NMarks) && (MarSel[j] < swp))
         {
            ggg[j][id] = fgl;
            ++j;
         }
         fgl = fgl2;
      }
      while (j < NMarks)
      {
         ggg[j][id] = fgl;
         ++j;
      }

   }

   return;
}

void check_fgl ()
{
   int   i, j;

   for (i = 0; i < pedsize; ++i)
      if (nghd[i].obs)
      {
         printf ("\n %d %s", nghd[i].index, nghd[i].self);
         for (j = 0; j < NMarks; ++j)
            printf (" %d", gg0[j][i]);
         printf ("   ");
         for (j = 0; j < NMarks; ++j)
            printf (" %d", gg1[j][i]);
      }
   printf ("\n");

   return;
}

void fgl_set (FILE * fp, int c_fgl)
{
   int   i;

   for (i = 0; i < c_fgl; ++i)
   {
      set_line (fp, gg0);
      set_line (fp, gg1);
   }

   if (DebugMain)
      check_fgl ();

   return;
}

void make_ibd_compnts (int j)
/* this one will take genes matrix at locus j, and determine components of
   the ibd graph among the individuals observed at this marker
   For each individual,  and each gg-node, we will determine a compnt index
   Also, give size of each compnt   */
{
   int   cc;                            /*  count of individuals assigned  */

   cc = 0;
/*   ******
      while cc < pedsize
      { find first unassigned individual
        if unobserved   ++cc ;  done for this one; set index to -1
        if observed:  increase # components by 1  ++nc;
                      add individual to components  ++cc
                      find his genes; assign his genes this compnt no.
              go down list of unassigned inds; if has either gene;
                   assign individual (++cc) , assign other gene(s)
            repeatwhile individuals get added
       }
      *********   */

   return;
}

void phase_ibd_compnts (int nc)
{
   /* go through individuals  */
   /*  if a homoz -- mark that component as phasable  */
   /* mark the allele types of the homozyotes gg-nodes */
   /* mark that individual with phase =1    */

/* Go through the phasable components: */

/*
    go though individuals
       find component k of indiv
       if components k is phasable;
       if individual  not yet phased
       if either gg-node assigned; assign the other
                assign phase  +/-1
              adding as above  --
              hence assign +/- 1 phase to all individuals
            in all phasable components
 */

   return;
}
