/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 * The following functions are called by proc_drop();
 * some are also called by proc_auto().
 */

/**
 * tests for markers or traits to be simulated.
 * checks for chromosome no(s) given
 */

int test_sim (void)
{
   SIM_PTR ps;

   if ((ps = first_sim_chrom_node))     /* checks for both named and unnamed  */
   {                                    /*  chromosomes                       */
      if (ps == 0 && ps->p_next)
      {
         err_typ = ERD_CHR;
         return FALSE;
      }
   }
   else if (!sim_traits)
   {                                    /* tests for simulate statement       */
      err_typ = ERD_SIM;
      return FALSE;
   }
   return TRUE;
}

/**
 * checks for tlocs specified for all traits to be selected or simulated.
 * marks the traits in the trait/tloc list as selected.
 */
int find_trait_tloc (void)
{
   int   tn, entry, entry2;
   TRAITTLOC_PTR p_tt;

   entry = FALSE;
   if (!(p_tt = nxt_tr_tl (&entry)))
   {
      err_typ = ERD_MTT;                /* if trait/tloc not specified, error */
      return FALSE;
   }

   entry2 = 0;                          /* for ea trait to be sim,            */
   while ((tn = nxt_sim_tr_no (&entry2)) && !err_typ)
   {
      entry = 0;                        /* find trait/tloc list               */
      while ((p_tt = nxt_tr_tl (&entry)) && p_tt->trait_no < tn)
         ;
      if (p_tt && p_tt->trait_no == tn) /* mark trait as selected             */
         p_tt->selectd = TRUE;
      else
      {
         err_typ = ERD_TTM;             /* if trait/tloc lst not found, error */
         sprintf (errnumb, "%d", tn);
         err_txt = errnumb;
      }
   }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/**
 * checks for tlocs specified for the inheritances selected.
 * marks the inheritances in the inher/tloc list as selected.
 */
int find_inher_tloc (void)
{
   int   ihn, entry, entry2;
   INHERTLOC_PTR p_it;

   entry = FALSE;
   if (!(p_it = nxt_inh_tl (&entry)))
   {
      err_typ = ERD_MIT;                /* if inher/tloc not specified, error */
      return FALSE;
   }

   entry2 = 0;                          /* for ea inher selected,             */
   while ((ihn = nxt_sim_tr_no (&entry2)) && !err_typ)
   {
      entry = 0;                        /* find inher/tloc list               */
      while ((p_it = nxt_inh_tl (&entry)) && p_it->inher_no < ihn)
         ;
      if (p_it && p_it->inher_no == ihn) /* mark inher as selected            */
         p_it->selectd = TRUE;
      else
      {
         err_typ = ERD_TIM;             /* if inher/tloc lst not found, error */
         sprintf (errnumb, "%d", ihn);
         err_txt = errnumb;
      }
   }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/**
 * looks for chromosome maps, or unlinked spec, for the requested tlocs
 * (tlocs in the trait/tloc list which are marked as selected).
 * marks these tlocs in the mapped or unlinked lists as selected.
 * checks for just one location for each tloc:
 *   if a tloc is not mapped or unlinked, it is an error.
 *   if a tloc is both mapped and unlinked, it is an error.
 *   if a tloc is mapped on more than one chromosome, it is an error.
 *
 *      for ea trait
          for ea sel trait
            for ea tloc for the trait
              for ea tloc in unlnkd list
                tloc nos match?
                  set unlnkd tloc as selctd
                  set found unlnkd tloc flag
              for ea gender
                for ea chrom tloc map
                  for ea mapd tloc in list
                    tloc nos match?
                      set mapd tloc as selctd
                      is found mapd tloc flag set?
                        n - set found mapd tloc flag
                            set chrom no. found to map chrom no.
                        y - is chrom no. found = map chrom no?
                            n - error         (mapd on more than 1 chrom)
            if !found unlnkd & !found lnkd  (map not present)
              error
            if found unlnkd & found lnkd    (both mapd & unlnkd)
              error
*/
int find_tr_tloc_map (void)
{
   TRAITTLOC_PTR p_tt;
   int   entry, i, tl_no;

   entry = FALSE;                       /* for ea trait/tloc in list:         */
   while ((p_tt = nxt_tr_tl (&entry)) && !err_typ) /* checks trait for selctd */
      if (p_tt->selectd)
         for (i = 0; i < p_tt->no_loci; ++i)
         {
            tl_no = *(p_tt->p_tloc + i); /* For each tloc for the trait:      */
            find_nxt_tl_map (tl_no);    /* Find its map and mark as selected. */
         }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/**
 * looks for the chromosome map for the specified tloc
 */
void find_nxt_tl_map (int tln)
{
   int   entry2, entry3, sx, foundu, foundl, chrom_no;
   MAP_PTR p_map;
   UDIST_PTR p_ud;
   DIST_PTR p_d;

   foundu = 0;                          /* first checks tloc in unlinked list */
   entry2 = FALSE;
   while ((p_ud = nxt_unln_dist (&entry2)) && !foundu)
   {
      if (p_ud->no == tln)              /* found tloc (unlinked)              */
      {
         p_ud->selectd = TRUE;
         ++foundu;
      }
   }

   chrom_no = 0;                        /* then checks for tloc mapped        */
   foundl = 0;
   for (sx = FEM; sx <= AVG; ++sx)      /* for each gender                    */
   {
      entry2 = FALSE;                   /* for each tloc map                  */
      while ((p_map = nxt_tl_map (&entry2, sx)) && !err_typ)
      {
         entry3 = FALSE;                /* checks tloc in list                */
         while ((p_d = nxt_dist (&entry3, p_map)) && !err_typ)
            if (p_d->no == tln)         /* found tloc (mapped)                */
            {
               p_d->selectd = TRUE;
               if (!foundl)
               {
                  foundl = TRUE;
                  chrom_no = p_map->chrom_no;
               }
               else if (chrom_no != p_map->chrom_no)
               {                        /* checks for tloc on > 1 chrom       */
                  err_typ = ERD_MCT;
                  sprintf (errnumb, "%d", tln);
                  err_txt = errnumb;
               }
            }
      }
   }
   if (!foundu && !foundl)              /* checks for not unlinked or mapped  */
   {
      err_typ = ERD_TNF;
      sprintf (errnumb, "%d", tln);
      err_txt = errnumb;
   }
   else if (foundu && foundl)           /* checks for both unlinked and mappd */
   {
      err_typ = ERD_TUL;
      sprintf (errnumb, "%d", tln);
      err_txt = errnumb;
   }
   return;
}

/**
 * looks for chromosome maps, or unlinked spec, for the requested inheritance
 * (tlocs in the inher/tloc list which are marked as selected).
 */
int find_inh_tloc_map (void)
{
   INHERTLOC_PTR p_it;
   int   entry, i, tl_no;

   entry = FALSE;
   while ((p_it = nxt_inh_tl (&entry)) && !err_typ) /* for ea inh/tloc list:  */
   {
      if (p_it->selectd)                /* checks inheritance for selected    */
      {
         for (i = 0; i < p_it->no_loci; ++i)
         {
            tl_no = *(p_it->p_tloc + i); /* For each tloc for the inheritance: */
            find_nxt_tl_map (tl_no);    /* Find its map and mark as selected. */
         }
      }
   }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/**
 * looks for chromosome maps, or unlinked specs, for the selected tlocs
 * (tlocs which are in the trait sim/sel list)
 */
int find_tloc_map (void)
{
   int   tl_no;

   tl_no = *(sim_traits->p_trait);
   find_nxt_tl_map (tl_no);             /* Find its map and mark as selected. */

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/**
 * for ea chromosome and gender, adds a list of the linked tlocs required
 */
int updt_sim_lnkd_tloc (void)
{
   int   j, sx, entry, entry2, entry3, nt;
   SIM_PTR p_sim;
   MAP_PTR p_map;
   DIST_PTR p_d;

   for (sx = FEM; sx <= AVG; ++sx)      /* finds linked tlocs to simulate     */
   {
      entry = FALSE;
      while ((p_map = nxt_tl_map (&entry, sx))) /* for each gender and chrom, */
      {
         nt = 0;
         entry2 = FALSE;
         while ((p_d = nxt_dist (&entry2, p_map)))
            if (p_d->selectd)
               ++nt;                    /* counts no. linked tlocs to sim     */

         if (nt)                        /* if tlocs required,                 */
         {
            entry3 = FALSE;
            while ((p_sim = nxt_sim_chrom (&entry3))
                   && p_sim->chrom_no < p_map->chrom_no)
               ;                        /* finds sim node for this chrom      */
            if (p_sim && p_map->chrom_no == p_sim->chrom_no)
            {
               if (!p_sim->no_tloc_sel) /* if tloc list not already set,      */
               {
                  p_sim->no_tloc_sel = nt; /* set no tlocs sel                */
                  p_sim->p_tloc = malloc (nt * sizeof (int));
                  alloc_chk (WHERE, p_sim->p_tloc);

                  entry2 = FALSE;
                  j = 0;
                  while ((p_d = nxt_dist (&entry2, p_map)))
                     if (p_d->selectd)  /* adds tloc nos to sim list          */
                     {
                        *(p_sim->p_tloc + j) = p_d->no;
                        ++j;
                     }
               }
            }
            else                        /* did not find sim markers for chrom */
            {
               err_typ = ERD_MSM;
               on_chrom (p_map->chrom_no);
            }
         }
      }
   }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/**
 * makes the sim list of the required unlinked tlocs
 */
int updt_sim_unln_tloc (void)
{
   int   j, entry, nt;
   UDIST_PTR p_u;

   nt = 0;
   entry = FALSE;
   while ((p_u = nxt_unln_dist (&entry)))
      if (p_u->selectd)                 /* counts no. unlinked tlocs to sim   */
         ++nt;

   if (nt)                              /* if tlocs required,                 */
   {
      sim_unlnkd_tlocs = malloc (sizeof (SIM_UNLNKD));
      alloc_chk (WHERE, sim_unlnkd_tlocs);
      sim_unlnkd_tlocs->no_tloc_sel = nt;
      sim_unlnkd_tlocs->p_tloc = malloc (nt * sizeof (int));
      alloc_chk (WHERE, sim_unlnkd_tlocs->p_tloc);

      j = 0;
      entry = FALSE;
      while ((p_u = nxt_unln_dist (&entry)))
      {
         if (p_u->selectd)              /* adds tloc nos to sim unlinked list */
         {
            *(sim_unlnkd_tlocs->p_tloc + j) = p_u->no;
            ++j;
         }
      }
   }
   return TRUE;
}

/**
 * checks the marker map for the selected chromosome.
 * tests for marker map present for this chromosome.
 * adds the no. of mapped markers to the sim node.
 * tests for consistent sex-specific marker maps.
 * if marker map is missing, checks that the map is not needed.
 * for selection of all markers, sets the number of markers
 * to the no. mapped.
 * for list of markers, checks marker no.'s against no. mapped.
 */
int test_mark_map (SIM_PTR p_sim, int *ss_map)
{
   int   sx, entry;
   MAP_PTR p_map;
   MAP_PTR p_m[3] = { NULL, NULL, NULL };

   for (sx = FEM; sx <= AVG; ++sx)      /* check for marker maps present      */
   {
      entry = FALSE;
      while ((p_map = nxt_mark_map (&entry, sx))
             && p_map->chrom_no < p_sim->chrom_no)
         ;
      if (p_map && p_map->chrom_no == p_sim->chrom_no) /* if found, adds no   */
      {                                 /* marks mapped to sim node           */
         p_sim->no_mark_map = p_map->no_loci;
         p_m[sx] = p_map;
         if (sx != AVG)                 /* checks for sex-specific map        */
            *ss_map = TRUE;
      }
   }
   if (p_m[FEM] && p_m[MAL] && (p_m[FEM]->no_loci != p_m[MAL]->no_loci))
      err_typ = ERD_NFM;
   if (p_m[MAL] && p_m[AVG] && (p_m[MAL]->no_loci != p_m[AVG]->no_loci))
      err_typ = ERD_NMA;
   if (p_m[AVG] && p_m[FEM] && (p_m[AVG]->no_loci != p_m[FEM]->no_loci))
      err_typ = ERD_NFA;
   if (err_typ)                         /* check for same no. markers for     */
   {                                    /* ea gender                          */
      for_chrom (p_sim->chrom_no);
      return FALSE;
   }

   if (!p_sim->no_mark_map)             /* **** no map is present        **** */
   {
      if (SS_MKALL == p_sim->mark_typ)  /* if all markers are selected        */
      {                                 /* assumes 1 marker                   */
         p_sim->no_mark_sel = 1;
         err_typ = ERA_1MA;             /* issues a warning                   */
      }
      else if (SS_MKLIST == p_sim->mark_typ) /* if marker list is selected    */
      {
         if (p_sim->no_mark_sel != 1)   /* accepts if only 1 marker           */
            err_typ = ERD_MMM;
         else if (*(p_sim->p_mark) != 1) /* and if marker no. is 1            */
            err_typ = ERA_MN1;
      }
      if (p_sim->no_mark_sel == 1)
         p_sim->no_mark_map = 1;
   }
   else                                 /* **** marker map is present    **** */
   {                                    /* for selection of all,              */
      if (SS_MKALL == p_sim->mark_typ)  /* sets no selected to no mapped      */
         p_sim->no_mark_sel = p_sim->no_mark_map;
      else if (SS_MKLIST == p_sim->mark_typ) /* for marker list, checks range */
         if (*(p_sim->p_mark + p_sim->no_mark_sel - 1) > p_sim->no_mark_map)
            err_typ = ERA_MNL;
   }

   if (err_typ)
   {
      for_chrom (p_sim->chrom_no);
      return FALSE;
   }
   else
      return TRUE;
}

/**
 * checks the tloc map for the chromosome.
 * tests for tloc map present for this chromosome.
 * checks sex-specific maps for consistency with ea other.
 * checks for marker no.'s within range.
 */
int test_tloc_map (SIM_PTR p_sim, int *ss_map)
{
   int   sx, j, entry;
   MAP_PTR p_map, p_sm[3] = { NULL, NULL, NULL };
   DIST_PTR p_d1, p_d2;

   if (!p_sim->no_tloc_sel)             /* checks for sim tlocs               */
      return TRUE;

   for (sx = FEM; sx <= AVG; ++sx)      /* checks for tloc map present        */
   {
      entry = FALSE;
      while ((p_map = nxt_tl_map (&entry, sx))
             && p_map->chrom_no < p_sim->chrom_no)
         ;
      if (p_map && p_map->chrom_no == p_sim->chrom_no)
      {
         p_sm[sx] = p_map;              /* if found, checks for sex-spec map  */
         if (sx != AVG)
            *ss_map = TRUE;
      }
   }
   if (!p_sm[AVG] && !p_sm[MAL] && !p_sm[FEM]) /* checks for map found        */
      err_typ = ERD_MTM;                /* find_tloc_map() checks for this E  */

   if (err_typ)
   {
      for_chrom (p_sim->chrom_no);
      return FALSE;
   }

   if (p_sm[FEM] && p_sm[MAL] && (p_sm[FEM]->no_loci != p_sm[MAL]->no_loci))
      err_typ = ERD_TFM;
   if (p_sm[MAL] && p_sm[AVG] && (p_sm[MAL]->no_loci != p_sm[AVG]->no_loci))
      err_typ = ERD_TMA;
   if (p_sm[AVG] && p_sm[FEM] && (p_sm[AVG]->no_loci != p_sm[FEM]->no_loci))
      err_typ = ERD_TFA;
   if (err_typ)                         /* checks for same no. tlocs for ea   */
   {                                    /* gender                             */
      for_chrom (p_sim->chrom_no);
      return FALSE;
   }

   for (sx = FEM; sx <= AVG; ++sx)      /* checks sex-specific maps for       */
   {                                    /* consistency                        */
      if (p_sm[sx])
      {
         p_d1 = p_sm[sx]->p_dist;
         p_d2 = NULL;
         if (p_sm[(sx + 1) % (AVG + 1)])
            p_d2 = p_sm[(sx + 1) % (AVG + 1)]->p_dist;

         for (j = 0; j < p_sm[sx]->no_loci; ++j)
         {
            if (p_sim->no_mark_map == 0) /* no marker map                     */
            {
               if (p_d1->map_mark > 1)  /* checks for marker no. 0 or 1       */
                  err_typ = ERD_MMX;
            }
            else if (p_d1->map_mark > p_sim->no_mark_map) /* checks for mark  */
               err_typ = ERD_MMX;       /* no. in tloc map in range           */
            else if (p_d2)              /* checks for tloc no. the same in    */
            {                           /* in sex-specific maps               */
               if (p_d1->no != p_d2->no)
                  err_typ = ERD_TID;
               else if (p_d1->map_mark != p_d2->map_mark) /* checks tloc      */
                  err_typ = ERD_MID;    /* the same locus in sex-spec maps    */
               ++p_d2;
            }
            ++p_d1;
         }
      }
   }
   if (err_typ)
   {
      for_chrom (p_sim->chrom_no);
      return FALSE;
   }
   else
      return TRUE;
}

/**
 *  checks for tloc already in tloc list;
 *  if not, adds tloc to list, in order of tloc no.
 */
void mak_tloc_list (int no, int lnkd)
{
   TLLIST_PTR cur_node, new_node;

   if (first_tllist_node)
   {
      cur_node = first_tllist_node;
      while (cur_node->p_next && cur_node->p_next->no <= no)
         cur_node = cur_node->p_next;

      if (cur_node && cur_node->no == no)
      {
         cur_node->lnkd = lnkd;
         ++(cur_node->n_us);
         return;
      }
   }

   new_node = malloc (sizeof (TLLIST));
   alloc_chk (WHERE, new_node);

   new_node->no = no;
   new_node->lnkd = lnkd;
   new_node->n_us = 1;

   if (!first_tllist_node || first_tllist_node->no > no)
   {
      new_node->p_next = first_tllist_node;
      first_tllist_node = new_node;
   }
   else
   {
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return;
}

/* Subroutine for deciding whether or not to warn the user that neither a     */
/* "set markers ... allele frequencies ..." statement nor a "set all markers  */
/* allele frequencies ..." statement will override the other.                 */

int test_locus_nodes (SIM_PTR p_sim)
{
   static int already_warned = FALSE;

   LOCUS_PTR all_node, indiv_node;

   if (!already_warned)                 /* There is no need to warn the user  */
      /*                                   of this more than one time.        */
   {
      all_node = all_mark_locus_node;
      indiv_node = first_mark_locus_node;

      /* We don't need a warning unless there are both all_mark_locus_node's  */
      /* and individual mark_locus_node's.                                    */

      if (all_node && indiv_node)
      {
         /* See if the selected chromosome is present in both linked lists.   */

         while (all_node && (all_node->chrom_no < p_sim->chrom_no))
            all_node = all_node->p_next;

         while (indiv_node && (indiv_node->chrom_no < p_sim->chrom_no))
            indiv_node = indiv_node->p_next;

         if (all_node && (all_node->chrom_no == p_sim->chrom_no) &&
             indiv_node && (indiv_node->chrom_no == p_sim->chrom_no))
         {
            already_warned = TRUE;      /* One warning is all that's needed.  */
            err_typ = ERA_NOO;
            return FALSE;
         }
      }
   }

   return TRUE;
}

/* Checks for allele frequencies present for the specified marker.            */
/* If not, checks for whether or not defaults are permitted.                  */
/* If present, checks for whether frequencies are to be normalized.           */

int test_mark_freq (SIM_PTR p_sim, int m_no)
{
   LOCUS_PTR p_locus;
   FREQ_PTR p_freq;

   set_alle_frq_deflts ();              /* Checks for normalizing, defaults.  */

   p_locus = first_mark_locus_node;

   if (p_locus == NULL)                 /* Looks for allele freqs, this locus */
   {
      p_locus = mark_freq_missing (p_sim, m_no);
   }
   else
   {
      while (p_locus->p_next && (p_locus->chrom_no < p_sim->chrom_no
                                 || (p_locus->chrom_no == p_sim->chrom_no
                                     && p_locus->no < m_no)))
         p_locus = p_locus->p_next;

      if (p_locus->chrom_no != p_sim->chrom_no
          || (p_locus->chrom_no == p_sim->chrom_no && p_locus->no != m_no))
      {                                 /* Test for allele freqs for marker.  */
         p_locus = mark_freq_missing (p_sim, m_no);
      }
   }

   if (err_typ)
   {
      if (alle_frq_deflts)              /* Checks for default freqs permitted */
      {
         p_freq = alloc_dflt_frq ();
         stor_dflt_frq (p_freq);
         p_locus = alloc_dflt_locus_node (p_sim->chrom_no, m_no, MARK_LOC);
         stor_dflt_locus_data (p_sim->chrom_no, m_no, p_locus, p_freq);
      }
   }
   else if (unnorm_frq (p_locus->p_freq_node->sum))
   {
      if (alle_frq_norm)                /* Checks for freqs to be normalized. */
         err_typ = ERD_MFN;

      else if (unnorm_frq_mx (alle_frq_norm, p_locus->p_freq_node->sum))
         err_typ = ERD_MFX;             /* No normalization, check sum > max. */

      for_chrom_mark (p_sim->chrom_no, m_no);
   }

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

int test_trait_vars (int t_no)
{
   int   no_addi = FALSE, no_resi = FALSE;
   double sum = 0.;
   VARI_PTR p_addi, p_resi;

   p_addi = first_addi_node;
   p_resi = first_resi_node;

   if (!p_addi)
      no_addi = TRUE;
   else
   {
      while (p_addi->p_next && p_addi->no < t_no)
         p_addi = p_addi->p_next;
      if (p_addi->no == t_no)
         sum = p_addi->vari;
      else
         no_addi = TRUE;
   }
   if (!p_resi)
      no_resi = TRUE;
   else
   {
      while (p_resi->p_next && p_resi->no < t_no)
         p_resi = p_resi->p_next;
      if (p_resi->no == t_no)
         sum = sum + p_resi->vari;
      else
         no_resi = TRUE;
   }
   /*                                    * stores default values             */
   if (no_addi)
   {
      p_addi = alloc_vari_node (t_no, AV);
      stor_dflt_vari_data (t_no, p_addi);
   }
   if (no_resi)
   {
      p_resi = alloc_vari_node (t_no, RV);
      stor_dflt_vari_data (t_no, p_resi);
   }
   if (no_addi && no_resi)
      err_typ = ERD_NAR;
   else if (sum == 0.)
      err_typ = ERD_SVR;
   else if (no_addi)
      err_typ = ERD_TAV;
   else if (no_resi)
      err_typ = ERD_TRV;

   if (err_typ)
   {
      sprintf (errnumb, "%d", t_no);
      err_txt = errnumb;
      return FALSE;
   }
   else
      return TRUE;
}

/* Checks for genotype means present for trait and tloc.                      */
/* If not, sets default values and issues a warning.                          */

int test_tr_tl_genotyp (int t_no, int tl_no)
{
   GENO_MEANS_PTR p_m;

   p_m = first_geno_means_node;         /* Checks for genotype means present. */

   if (!p_m)
      err_typ = ERD_NGM;
   else                                 /* Looks for means for trait at tloc. */
   {
      while (p_m->p_next
             && (p_m->trait_no < t_no
                 || (p_m->trait_no == t_no && p_m->tloc_no < tl_no)))
         p_m = p_m->p_next;

      if ((p_m && p_m->trait_no != t_no) || p_m->tloc_no != tl_no)
         err_typ = ERD_NGM;
   }

   if (err_typ)
   {
      for_trait_at_tloc (t_no, tl_no);
      p_m = alloc_geno_means_node (t_no, tl_no);

      /* Store default values: 3 genotype means set to 0.0 each.              */
      nx_par = 3;
      alloc_geno_means_array (p_m);
      stor_dflt_geno_mean_dat (t_no, tl_no, p_m);

      return FALSE;
   }
   else
      return TRUE;
}

/**
 * checks for allele frequencies proesent for specified tloc.
 * if not, checks for defaults permitted.
 * if present, checks for frequencies to be normalized.
 */
int test_tloc_freq (int t_no)
{
   LOCUS_PTR p_locus;
   FREQ_PTR p_freq;

   set_alle_frq_deflts ();              /* checks for normalizing, defaults   */

   p_locus = first_tloc_locus_node;

   if (!p_locus)                        /* looks for allele freqs, this trait */
   {
      if (alle_frq_deflts)
         err_typ = ERD_NTL;
      else
         err_typ = ERD_TFR;
   }
   else
   {
      while (p_locus->p_next && p_locus->no < t_no)
         p_locus = p_locus->p_next;

      if (p_locus->no != t_no)
      {
         if (alle_frq_deflts)
            err_typ = ERD_NTL;
         else
            err_typ = ERD_TFR;
      }
   }
   if (err_typ)
   {
      sprintf (errnumb, "%d", t_no);
      err_txt = errnumb;

      if (alle_frq_deflts)              /* checks for default freqs permitted */
      {
         p_freq = alloc_dflt_frq ();
         stor_dflt_frq (p_freq);
         p_locus = alloc_dflt_locus_node (NO_CH_NO, t_no, TLOC_LOC);
         stor_dflt_locus_data (NO_CH_NO, t_no, p_locus, p_freq);
      }
   }
   else if (unnorm_frq (p_locus->p_freq_node->sum))
   {
      if (alle_frq_norm)                /* checks for freqs to be normalized  */
         err_typ = ERD_TFN;

      else if (unnorm_frq_mx (alle_frq_norm, p_locus->p_freq_node->sum))
         err_typ = ERD_TFX;

      sprintf (errnumb, "%d", t_no);
      err_txt = errnumb;
   }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/* Handles the situation where no locus_node for the given marker number has  */
/* been located.  If a "set all markers allele frequencies" parameter state-  */
/* ment was given by the user, those frequencies will be used for the given   */
/* marker number.  Otherwise, construct an appropriate error message.         */

LOCUS_PTR mark_freq_missing (SIM_PTR p_sim, int m_no)
{
   LOCUS_PTR source_node, dest_node;
   FREQ_PTR freq_node;

   /* See if a "set all markers allele frequencies" statement was provided by */
   /* the user for the selected chromosome.                                   */

   source_node = all_mark_locus_node;
   while (source_node && (source_node->chrom_no < p_sim->chrom_no))
      source_node = source_node->p_next;

   if (source_node && (source_node->chrom_no == p_sim->chrom_no))
   {
      /* Set up an individual locus_node for the specified marker number.     */

      freq_node = source_node->p_freq_node;
      nx_par = source_node->no_alleles;

      if (p_sim->chrom_no > 0)
      {
         j_par[0] = p_sim->chrom_no;
         dest_node = alloc_locus_node (m_no, CH, MARK_LOC);
         stor_locus_data (m_no, dest_node, freq_node, CH, MARK_LOC);
      }
      else
      {
         dest_node = alloc_locus_node (m_no, NCH, MARK_LOC);
         stor_locus_data (m_no, dest_node, freq_node, NCH, MARK_LOC);
      }
      ++dest_node->p_freq_node->no_loci;
   }

   else
   {
      if (alle_frq_deflts)
         err_typ = ERD_NML;
      else
         err_typ = ERD_MFR;
      for_chrom_mark (p_sim->chrom_no, m_no);
   }

   return dest_node;
}

void for_chrom (int chr_no)
{
   if (chr_no > 0)
      sprintf (errnumb, "for chromosome %d", chr_no);
   else
      errnumb[0] = '\0';
   err_txt = errnumb;
   return;
}

void on_chrom (int chr_no)
{
   if (chr_no > 0)
      sprintf (errnumb, "on chromosome %d", chr_no);
   else
      errnumb[0] = '\0';
   err_txt = errnumb;
   return;
}

void for_chrom_mark (int chr_no, int m_no)
{
   if (chr_no > 0)
      sprintf (errnumb, "for chromosome %d marker %d", chr_no, m_no);
   else
      sprintf (errnumb, "for marker %d", m_no);
   err_txt = errnumb;
   return;
}

#if 0
void for_chrom_tloc (int chr_no, int tl_no)
{
   if (chr_no > 0)
      sprintf (errnumb, "for chromosome %d tloc %d", chr_no, tl_no);
   else
      sprintf (errnumb, "for tloc %d", tl_no);
   err_txt = errnumb;
   return;
}
#endif

/**
 * formats trait and tloc numbers.
 */
void for_trait_at_tloc (int tr_no, int tl_no)
{
   sprintf (errnumb, "for trait %d at tloc %d", tr_no, tl_no);
   err_txt = errnumb;
   return;
}

PRNTMAP_PTR add_prntmap (SIM_PTR p_sim, int sx)
{
   PRNTMAP_PTR p_pmap, p_cur;

   p_pmap = malloc (sizeof (PRNTMAP));  /* makes prntmap node for this       */
   alloc_chk (WHERE, p_pmap);           /* chromosome                        */

   p_pmap->chrom_no = p_sim->chrom_no;
   p_pmap->no_mark_sel = p_sim->no_mark_sel;
   p_pmap->no_mark_map = p_sim->no_mark_map;
   p_pmap->no_tloc_sel = p_sim->no_tloc_sel;
   p_pmap->p_locus = NULL;
   p_pmap->p_next = NULL;

   if (!first_prntmap_node[sx])         /* puts node as first in list        */
      first_prntmap_node[sx] = p_pmap;
   else
   {
      p_cur = first_prntmap_node[sx];   /* or adds node to end of list       */
      while (p_cur->p_next)
         p_cur = p_cur->p_next;
      p_cur->p_next = p_pmap;
   }
   return p_pmap;
}

PRDIST_PTR add_locus_node (void)
{
   PRDIST_PTR p_new;

   p_new = malloc (sizeof (PRDIST));    /* allocs and inits prntmap_locus     */
   alloc_chk (WHERE, p_new);            /* node                               */

   p_new->locus_typ = 0;
   p_new->no = 0;
   p_new->map_mark = 0;
   p_new->selectd = 0;
   p_new->sel_map_mark = 0;
   p_new->dist = 0.;
   p_new->sel_dist = 0.;
   p_new->m1_dist = 0.;
   p_new->loc_dist = 0.;
   p_new->p_prev = NULL;
   p_new->p_next = NULL;

   return p_new;
}

PRDIST_PTR add_only_mark_locus (PRNTMAP_PTR p_pmap)
{
   PRDIST_PTR p_new;

   p_new = add_locus_node ();           /* makes prntmap_locus node for mark  */

   p_new->no = 1;                       /* the marker is unmapped             */
   p_new->locus_typ = MARK_LOC;

   p_pmap->p_locus = p_new;

   return p_new;
}

PRDIST_PTR add_mark_locus (PRNTMAP_PTR p_pmap, DIST_PTR p_dist)
{
   static double d1;
   PRDIST_PTR p_new, p_cur;

   p_new = add_locus_node ();           /* makes prntmap_locus node for mark  */

   if (p_dist->no == 1)
      d1 = 0.;
   p_new->locus_typ = MARK_LOC;
   p_new->no = p_dist->no;
   p_new->map_mark = p_dist->map_mark;
   p_new->dist = p_dist->dist;
   d1 += p_dist->dist;
   p_new->m1_dist = d1;

   if (!p_pmap->p_locus)                /* puts node at start of list         */
      p_pmap->p_locus = p_new;
   else
   {
      p_cur = p_pmap->p_locus;          /* or adds node to end of list        */
      while (p_cur->p_next)
         p_cur = p_cur->p_next;
      p_cur->p_next = p_new;
      p_new->p_prev = p_cur;
   }

   return p_new;
}

PRDIST_PTR add_tloc_locus (PRNTMAP_PTR p_pmap, DIST_PTR p_dist)
{
   PRDIST_PTR p_new, p_cur;

   p_new = add_locus_node ();           /* makes prntmap_locus node for tloc  */

   p_new->no = p_dist->no;
   p_new->map_mark = p_dist->map_mark;
   p_new->locus_typ = TLOC_LOC;
   p_new->dist = p_dist->dist;

   if (p_dist->map_mark == 0)           /* tloc precedes 1st marker           */
   {
      if (p_pmap->p_locus->locus_typ == MARK_LOC
          || (p_pmap->p_locus->locus_typ == TLOC_LOC
              && p_pmap->p_locus->dist < p_dist->dist))
      {
         p_new->p_next = p_pmap->p_locus; /* puts tloc at start of chrom      */
         p_pmap->p_locus->p_prev = p_new;
         p_pmap->p_locus = p_new;
      }
      else
      {
         p_cur = p_pmap->p_locus;
         while (p_cur->p_next
                && p_cur->p_next->locus_typ == TLOC_LOC
                && p_cur->p_next->dist > p_dist->dist)
            p_cur = p_cur->p_next;

         p_cur->p_next->p_prev = p_new; /* puts tloc after start              */
         p_new->p_next = p_cur->p_next;
         p_new->p_prev = p_cur;
         p_cur->p_next = p_new;
      }
   }
   else
   {                                    /* tloc follows 1st marker            */
      p_cur = p_pmap->p_locus;
      while (p_cur->p_next
             && (p_cur->p_next->locus_typ == TLOC_LOC
                 || (p_cur->p_next->locus_typ == MARK_LOC
                     && p_cur->p_next->no <= p_dist->map_mark)))
         p_cur = p_cur->p_next;

      if (p_cur->p_next)
         p_cur->p_next->p_prev = p_new;
      p_new->p_next = p_cur->p_next;    /* inserts tloc in list or adds       */
      p_new->p_prev = p_cur;            /* to end                             */
      p_cur->p_next = p_new;
   }

   return p_new;
}

/**
 * sets mapping information for "selected" markers.
 * determines:
 *    whether this marker is selected
 *    the selected reference marker for mapping
 *    the dist from the selected marker
 */
void sel_mark_locus (SIM_PTR p_sel, PRDIST_PTR p_pl, int *p_entry)
/*
 * for this gender and this locus,  adds info to prntmap_locus structures
 */
{
   PRDIST_PTR p_cur;
   int   i;
   static int s_i;                      /* relative selected no.              */
   static int s_no;                     /* selected marker no.                */
   static int s_no1;                    /* 1st selected marker no.            */

   if (!*p_entry)                       /* tests for start of marker list     */
   {
      *p_entry = TRUE;
      s_i = 0;
      if (p_sel->mark_typ != SS_MKALL)
      {
         s_no1 = *(p_sel->p_mark);
         s_no = s_no1;
      }
   }

   if (p_sel->mark_typ == SS_MKALL)     /* if all markers selected, copies    */
   {                                    /* marker & dist for mapping          */
      p_pl->selectd = ++s_i;
      p_pl->sel_map_mark = p_pl->map_mark;
      p_pl->sel_dist = p_pl->dist;

      return;
   }

   p_pl->sel_map_mark = s_no;           /* if selectn, sets marker for mappng */
   p_pl->selectd = 0;                   /* sets selectn/non-selectn indicator */
   for (i = 0; i < p_sel->no_mark_sel; ++i)
      if (p_pl->no == *(p_sel->p_mark + i))
      {
         p_pl->selectd = ++s_i;
         s_no = p_pl->no;
      }

   if (p_pl->no < s_no1)                /* marker precedes 1st selectd markr: */
      ;                                 /* postpones setting sel dist         */

   else if (p_pl->no == s_no1)          /* 1st selected marker:               */
   {                                    /* sets sel dist to 0.                */
      p_pl->sel_dist = 0.;
      p_pl->sel_map_mark = 0;

      if (p_pl->no > 1)                 /* goes back and sets sel dists for   */
      {                                 /* markers preceding this one         */
         p_cur = p_pl;
         while ((p_cur = p_cur->p_prev))
            p_cur->sel_dist = p_cur->p_next->sel_dist + p_cur->p_next->dist;
      }
   }
   else if (p_pl->no > s_no1)           /* marker follows 1st sel mark:       */
   {
      if (p_pl->p_prev->selectd)        /* if last marker was selected, sets  */
         p_pl->sel_dist = p_pl->dist;   /* sel dist to map dist;  if not sel, */
      else                              /* sets sel dist to map dist + prev   */
         p_pl->sel_dist = p_pl->dist + p_pl->p_prev->sel_dist; /* sel dist    */
   }

   return;
}

/**
 * checks marker to trait distance.
 * should not be > marker to marker distance.
 */
int test_tloc_dist (PRNTMAP_PTR p_pmap, PRDIST_PTR p_loc)
{
   PRDIST_PTR p_cur;

   if (p_loc == p_pmap->p_locus)        /*checks for 1st locus                */
      return TRUE;

   else if (p_loc->locus_typ == MARK_LOC) /*checks for marker locus           */
      return TRUE;

   else if (p_loc->map_mark == 0)       /*checks for no preceding mark locus  */
      return TRUE;

   else if (!p_loc->p_next)             /*checks for no following mark locus  */
      return TRUE;

   else
   {
      p_cur = p_loc->p_next;

      while (p_cur->p_next && p_cur->locus_typ != MARK_LOC)
         p_cur = p_cur->p_next;

      if (p_cur->locus_typ == MARK_LOC  /*tests marker to tloc dist           */
          && p_cur->dist < p_loc->dist)
      {
         err_typ = ERD_MTD;
         sprintf (errnumb, "%d", p_loc->no);
         err_txt = errnumb;
      }
   }
   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/**
 * finds distance between preceding locus & current locus.
 */
void fix_locus_dist (PRNTMAP_PTR p_pmap, PRDIST_PTR p_loc)
{
   if (p_loc == p_pmap->p_locus)        /*if first locus, dist is 0.          */
      p_loc->loc_dist = 0.;

   /*                                    * if tloc precedes 1st marker,       */
   /*                                    * dist is preceding tloc to marker   */
   /*                                    * less tloc to marker dist           */
   else if (p_loc->locus_typ == TLOC_LOC && p_loc->map_mark == 0)
      p_loc->loc_dist = p_loc->p_prev->dist - p_loc->dist;

   /*                                    * if marker is 1st marker,           */
   /*                                    * dist is preceding tloc to marker   */
   /*                                    * dist                               */
   else if (p_loc->locus_typ == MARK_LOC && p_loc->selectd == 1)
      p_loc->loc_dist = p_loc->p_prev->dist;

   /*                                    * if marker or tloc follows marker   */
   /*                                    * dist is ok                         */
   else if (p_loc->p_prev->locus_typ == MARK_LOC)
      p_loc->loc_dist = p_loc->dist;

   /*                                    * if marker follows tloc,            */
   /*                                    * dist is preceding marker to        */
   /*                                    * marker dist less preceding         */
   /*                                    * marker to tloc dist                */
   else if (p_loc->locus_typ == MARK_LOC
            && p_loc->p_prev->locus_typ == TLOC_LOC)
      p_loc->loc_dist = p_loc->dist - p_loc->p_prev->dist;

   /*                                    * if tloc follows trait,             */
   /*                                    * dist is preceding marker to tloc   */
   /*                                    * dist less preceding marker to      */
   /*                                    * preceding tloc dist                */
   else if (p_loc->locus_typ == TLOC_LOC
            && p_loc->p_prev->locus_typ == TLOC_LOC)
      p_loc->loc_dist = p_loc->dist - p_loc->p_prev->dist;

   return;
}

/* Checks that distance between preceding locus and current locus is greater  */
/* than or equal to some minimum distance.                                    */

int test_loc_mindist (PRDIST_PTR p_ploc, double mindist)
{
   /* No further check is needed if this is the first selected marker.        */
   if (p_ploc->selectd == 1)
      return TRUE;

   /* The distance between adjacent selected markers must be greater than 0.0 */
   /* cM.                                                                     */
   if (p_ploc->sel_dist == 0.0)
      err_typ = ERA_ZMD;

   /* For lm_map the distance between adjacent markers may not be less than   */
   /* the minimum distance passed as mindist to this subroutine.              */
   else if (LM_MAP && p_ploc->sel_dist < mindist)
   {
      sprintf (errnumb, "%.2f", mindist);
      err_txt = errnumb;
      err_typ = ERA_SMD;
   }

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/**
 * prints the chromosome map.
 * ea chromosome segment shows the marker and tloc positions, and
 * distances between the adjacent loci (cM)
 * (a segment consists of as many loci as will fit on the page)
 */
void prnt_map (int sxsp_mp)
{
   int   sex, no_loci;
   static int entry[3] = { 0, 0, 0 };
   PRDIST_PTR p_cur[3];
   PRNTMAP_PTR p_pmap;

   for (sex = FEM; sex <= AVG; ++sex)
      if ((p_pmap = nxt_prntmap (&entry[sex], sex)))
         p_cur[sex] = p_pmap->p_locus;

   if (!p_pmap)                         /* unexpectedly finds no map          */
      return;

/*                                       * first prints the heading           */
   if (p_pmap->no_mark_sel <= 1 && p_pmap->no_tloc_sel <= 0)
   {
      if (PrintLevel >= 4)
      {
         if (p_pmap->chrom_no != 0)
            fprintf (stdout, "\n No map for chromosome %d\n\n\n",
                     p_pmap->chrom_no);
         else
            fprintf (stdout, "\n No chromosome map\n\n\n");
      }

      return;
   }

   if (PrintLevel >= 4)
      fprintf (stdout, "\n Inter-locus distances in cM,"
               " using Haldane map function:\n\n");

   while (p_cur[AVG] && p_cur[AVG]->p_next) /* inits next chrom segment lines */
   {
      while (p_cur[AVG] && p_cur[AVG]->locus_typ == MARK_LOC
             && !p_cur[AVG]->selectd)
         for (sex = FEM; sex <= AVG; ++sex)
            p_cur[sex] = p_cur[sex]->p_next;

      if (p_cur[AVG]->locus_typ == TLOC_LOC) /* prints 1st loc on the chrom   */
         prntmap_tl0 (p_cur[AVG]->no);
      else
         prntmap_mrk0 (p_cur[AVG]->no);

      for (sex = FEM; sex <= AVG; ++sex)
         p_cur[sex] = p_cur[sex]->p_next;
      no_loci = 1;

      while (p_cur[AVG] && no_loci < max_line_loc) /* adds next locus to segm */
      {
         if (p_cur[AVG]->locus_typ == TLOC_LOC)
            prntmap_tl (no_loci, p_cur[AVG]->no);
         else
            prntmap_mrk (no_loci, p_cur[AVG]->no);

         for (sex = FEM; sex <= AVG; ++sex) /* adds next avg or FEM, MAL dist */
         {
            if ((sxsp_mp && (sex == FEM || sex == MAL))
                || (!sxsp_mp && sex == AVG))
               prntmap_dist (sex, no_loci, p_cur[sex]->loc_dist);
            p_cur[sex] = p_cur[sex]->p_next;
            while (p_cur[sex] && p_cur[sex]->locus_typ == MARK_LOC
                   && !p_cur[sex]->selectd)
               p_cur[sex] = p_cur[sex]->p_next;
         }
         ++no_loci;
      }
      prntmap_line (no_loci, sxsp_mp);  /* Print line for this map segment.   */

      for (sex = FEM; sex <= AVG; ++sex)
      {
         if (p_cur[sex])
            p_cur[sex] = p_cur[sex]->p_prev;
         while (p_cur[sex] && p_cur[sex]->locus_typ == MARK_LOC
                && !p_cur[sex]->selectd)
            p_cur[sex] = p_cur[sex]->p_prev;
      }
   }
   return;
}

void prntmap_mrk0 (int m_no)
{
   sprintf (trait_line, "    ");
   sprintf (trait_tic_marks, " -");
   sprintf (mark_tic_marks, " +");
   sprintf (mark_line, "M%d    ", m_no);

   return;
}

void prntmap_tl0 (int t)
{
   sprintf (trait_line, "T%d  ", t);
   sprintf (trait_tic_marks, " +");
   sprintf (mark_tic_marks, " -");
   sprintf (mark_line, "      ");

   return;
}

void prntvmap_tl0 (int f)
{
   if (f)
      sprintf (trait_line, " T* ");
   else
      sprintf (trait_line, " T  ");
   sprintf (trait_tic_marks, " +");
   sprintf (mark_tic_marks, " -");
   sprintf (mark_line, "      ");

   return;
}

void prntmap_tl0b (void)
{
   sprintf (trait_line, "    ");
   sprintf (trait_tic_marks, " +");
   sprintf (mark_tic_marks, " -");
   sprintf (mark_line, "      ");

   return;
}

void prntmap_tl (int n_loc, int t)
{
   sprintf (trait_line + n_loc * loc_spac - 3, "   T%d  ", t);
   sprintf (trait_tic_marks + n_loc * loc_spac - 5, "------+  ");
   sprintf (mark_tic_marks + n_loc * loc_spac - 5, "-------  ");
   sprintf (mark_line + n_loc * loc_spac - 1, "       ");

   return;
}

void prntvmap_tl (int n_loc, int f)
{
   if (f)
      sprintf (trait_line + n_loc * loc_spac - 3, "    T* ");
   else
      sprintf (trait_line + n_loc * loc_spac - 3, "    T  ");
   sprintf (trait_tic_marks + n_loc * loc_spac - 5, "------+  ");
   sprintf (mark_tic_marks + n_loc * loc_spac - 5, "-------  ");
   sprintf (mark_line + n_loc * loc_spac - 1, "       ");

   return;
}

void prntmap_mrk (int n_loc, int m_no)
{
   sprintf (trait_line + n_loc * loc_spac - 3, "       ");
   sprintf (trait_tic_marks + n_loc * loc_spac - 5, "-------  ");
   sprintf (mark_tic_marks + n_loc * loc_spac - 5, "------+  ");
   sprintf (mark_line + n_loc * loc_spac - 1, " M%d    ", m_no);

   return;
}

void prntmap_dist (int sex, int n_loc, double d)
{
   sprintf (dist_line[sex] + n_loc * loc_spac - 5, "%5.1f    ", d);

   return;
}

/* Prints the next segment of the chromosome map.  Before printing the next   */
/* segment, remove all trailing blank characters from the end of the output   */
/* lines for the chromosome map.                                              */

void prntmap_line (int n_loc, int sxsp)
{
   if (PrintLevel >= 4)
   {
      trim_blanks (trait_line);
      trim_blanks (trait_tic_marks);
      if (!sxsp)
         trim_blanks (dist_line[AVG]);
      else
      {
         trim_blanks (dist_line[FEM]);
         trim_blanks (dist_line[MAL]);
      }
      trim_blanks (mark_tic_marks);
      trim_blanks (mark_line);

      if (strchr (trait_line, 'T'))
         fprintf (stdout, "%s\n", trait_line);
      fprintf (stdout, "%s\n", trait_tic_marks);
      if (!sxsp)
         fprintf (stdout, "%s\n", dist_line[AVG]);
      else
      {
         fprintf (stdout, "%s\n", dist_line[FEM]);
         fprintf (stdout, "%s\n", dist_line[MAL]);
      }
      fprintf (stdout, "%s\n", mark_tic_marks);
      if (strchr (mark_line, 'M'))
         fprintf (stdout, "%s\n", mark_line);
      fprintf (stdout, "\n");
      fflush (stdout);
   }

   return;
}

/**
 * Prints the marker frequencies for this chromosome and gender.
 */
void prnt_mark_freqs (SIM_PTR p_sim)
{
   LOCUS_PTR cur_mark;

   char  prnt_lin[81] = "";
   double val;

   int   col_width;
   int   tot_cols;
   int   cols_used;
   int   m_ndx;
   int   ix;

   if (PrintLevel >= 5)
   {
      if (!first_mark_locus_node)
         return;

      cur_mark = first_mark_locus_node; /* Finds the alleles for the chrom.   */
      while (cur_mark->p_next && cur_mark->chrom_no < p_sim->chrom_no)
         cur_mark = cur_mark->p_next;

      if (cur_mark->chrom_no != p_sim->chrom_no)
         return;

      fprintf (stdout, " Marker allele frequencies:\n\n");

      /* Calculate how wide an output column needs to be to leave enough room */
      /* for the largest marker number mapped on the current chromosome.  The */
      /* column width is then used to calculate how many output columns can   */
      /* fit in a single output line.  If the column width is 5 the number of */
      /* columns is 15; if the column width is 6 the number of columns is 13; */
      /* and if the column width is 7 the number of columns is 11.            */

      sprintf (prnt_lin, "%d", p_sim->no_mark_map);
      col_width = max (strlen (prnt_lin) + 2, 5);
      tot_cols = 78 / col_width;

      cols_used = 0;
      m_ndx = 0;

      while (cur_mark && cur_mark->chrom_no == p_sim->chrom_no)
      {
         if ((m_ndx < p_sim->no_mark_sel)
             && (!p_sim->p_mark
                 || (p_sim->p_mark
                     && cur_mark->no == *(p_sim->p_mark + m_ndx))))
         {
            /* Print the allele frequencies for the selected markers.         */
            if (cols_used > 0)
            {
               if (cur_mark->no_alleles > tot_cols - cols_used - 2)
               {
                  fprintf (stdout, "%s\n", prnt_lin);
                  cols_used = 0;
               }
               else
               {
                  sprintf (prnt_lin + col_width * cols_used, "       ");
                  ++cols_used;
               }
            }
            sprintf (prnt_lin + col_width * cols_used, " M%d    ",
                     cur_mark->no);
            ++cols_used;

            ix = 0;
            while (ix < cur_mark->no_alleles)
            {
               if (cols_used >= tot_cols)
               {
                  fprintf (stdout, "%s\n", prnt_lin);
                  cols_used = 0;
               }
               val = *(cur_mark->p_freq_node->p_freq + ix);
               if (alle_frq_norm)
                  val /= cur_mark->p_freq_node->sum;
               sprintf (prnt_lin + col_width * cols_used, "%*.3f", col_width,
                        val);
               ++cols_used;
               *(prnt_lin + col_width * cols_used - 5) = ' ';
               ++ix;
            }
            ++m_ndx;
         }
         cur_mark = cur_mark->p_next;
      }

      if (cols_used > 0)
         fprintf (stdout, "%s\n", prnt_lin);
      fprintf (stdout, "\n\n");
      fflush (stdout);
   }

   return;
}

/**
 * Prints allele frequencies for the tlocs to be simulated
 * for the specified chromosome.
 */
void prnt_tloc_freqs (PRNTMAP_PTR p_pmap)
{
   char  prnt_lin[81] = "";
   int   i;

   PRDIST_PTR to_do;

   if (!p_pmap->no_tloc_sel)
      return;

   if (PrintLevel >= 2)
      fprintf (stdout, " Tloc allele frequencies:\n\n");

   to_do = p_pmap->p_locus;

   for (i = 0; i < p_pmap->no_tloc_sel; ++i)
   {                                    /* Finds next tloc number in map.     */
      while (to_do->locus_typ != TLOC_LOC && to_do->p_next)
         to_do = to_do->p_next;
      fmt_tloc_freqs (prnt_lin, i, to_do->no); /* Formats the frequencies.    */

      if (PrintLevel >= 2)
      {
         if (i + 1 == p_pmap->no_tloc_sel || !((i + 1) % 4))
            fprintf (stdout, "%s\n", prnt_lin); /* Prints the line.           */
         else
            sprintf (prnt_lin + (i % 4) * 20 + 15, "     ");
      }

      if (to_do->p_next)
         to_do = to_do->p_next;
   }
   if (PrintLevel >= 2)
   {
      fprintf (stdout, "\n\n");
      fflush (stdout);
   }

   return;
}

/**
 * Prints allele frequencies for the unlinked tlocs to be simulated
 * (assumes diallelic tlocs).
 */
void prnt_unltloc_freqs (void)
{
   char  prnt_lin[81] = "";
   int   i;

   if (!sim_unlnkd_tlocs)
      return;

   if (PrintLevel >= 2)
      fprintf (stdout, " Tloc allele frequencies:\n\n");

   for (i = 0; i < sim_unlnkd_tlocs->no_tloc_sel; ++i)
   {
      fmt_tloc_freqs (prnt_lin, i, *(sim_unlnkd_tlocs->p_tloc + i));

      if (PrintLevel >= 2)
      {
         if (i + 1 == sim_unlnkd_tlocs->no_tloc_sel || !((i + 1) % 4))
            fprintf (stdout, "%s\n", prnt_lin); /* Prints the line.           */
         else
            sprintf (prnt_lin + (i % 4) * 20 + 15, "     ");
      }
   }
   if (PrintLevel >= 2)
   {
      fprintf (stdout, "\n\n");
      fflush (stdout);
   }

   return;
}

/**
 * Prints the inheritance/tloci list.
 */
void prnt_inh_tl_list (void)
{
   int   entry, entry2, no, lno;

   if (PrintLevel >= 2)
   {
      fprintf (stdout, " Inheritance  tloc\n\n");

      entry = FALSE;
      while ((no = nxt_sim_tr_no (&entry)))
      {
         fprintf (stdout, "       %5d", no);
         entry2 = FALSE;
         while ((lno = nxt_inh_tl_no (&entry2, no)))
            fprintf (stdout, " %5d ", lno);
      }
      fprintf (stdout, "\n\n\n");
      fflush (stdout);
   }

   return;
}

/**
 * Prints names of the unlinked tlocs to be simulated.
 */
void prnt_unltloc_list (void)
{
   char  prnt_lin[81] = "";
   int   i;

   if (!sim_unlnkd_tlocs)
      return;

   if (PrintLevel >= 2)
   {
      for (i = 0; i < sim_unlnkd_tlocs->no_tloc_sel; ++i)
      {
         sprintf (prnt_lin + (i % 10) * 7, " T%d    ",
                  *(sim_unlnkd_tlocs->p_tloc + i));

         if (i + 1 == sim_unlnkd_tlocs->no_tloc_sel || !((i + 1) % 10))
            fprintf (stdout, "%s\n", prnt_lin); /* Prints the line.           */
      }
      fprintf (stdout, "\n\n");
      fflush (stdout);
   }
}

/**
 * Prints the trait list.
 */
void prnt_tr_list (void)
{
   int   entry, no;

   if (PrintLevel >= 2)
   {
      fprintf (stdout, " Trait\n\n");

      entry = FALSE;
      while ((no = nxt_sim_tr_no (&entry)))
         fprintf (stdout, " %5d\n", no);
      fprintf (stdout, "\n");
      fflush (stdout);
   }

   return;
}

/**
 * Prints the trait/tloci list.
 */
void prnt_tr_tl_list (void)
{
   int   entry, entry2, no, lno;

   if (PrintLevel >= 2)
   {
      fprintf (stdout, " Trait  tloc\n\n");

      entry = FALSE;
      while ((no = nxt_sim_tr_no (&entry)))
      {
         fprintf (stdout, " %5d", no);
         entry2 = FALSE;
         while ((lno = nxt_tr_tl_no (&entry2, no)))
            fprintf (stdout, " %5d\n", lno);
      }
      fprintf (stdout, "\n");
      fflush (stdout);
   }

   return;
}

/**
 * Prints the genotype means for the traits at the tloci.
 */
void prnt_tr_tl_genotyp (void)
{
   int   entry, entry2, no, lno;

   if (PrintLevel >= 2)
   {
      fprintf (stdout, " Trait  tloc                   genotype means\n");
      fprintf (stdout, "                          (1 1)     (1 2)     (2 2)\n");
      fprintf (stdout, "\n");

      entry = FALSE;
      while ((no = nxt_sim_tr_no (&entry)))
      {
         entry2 = FALSE;
         while ((lno = nxt_tr_tl_no (&entry2, no)))
            fmt_trtl_geno_means (no, lno); /* Format genotype means at tlocs. */
      }
      fprintf (stdout, "\n");
      fflush (stdout);
   }

   return;
}

/**
 * Prints the additive and residual variances for the traits.
 */
void prnt_trait_vari (void)
{
   int   entry, no;
   char  prnt_lin[81];

   if (PrintLevel >= 2)
   {
      fprintf (stdout, "\n Trait          variances\n");
      fprintf (stdout, "            additive  residual\n\n");

      entry = FALSE;
      while ((no = nxt_sim_tr_no (&entry)))
      {
         /* Formats trait additive and residual variances.                    */
         fmt_trait_vari (prnt_lin, no);
         fprintf (stdout, "%s\n", prnt_lin);
      }
      fprintf (stdout, "\n");
      fflush (stdout);
   }

   return;
}

/**
 * Formats the allele frequencies for the tloc.
 */
void fmt_tloc_freqs (char *p_lin, int n, int t_no)
{
   double val;

   LOCUS_PTR cur_tloc;
   /*                                    * Finds locus node for this trait.   */
   cur_tloc = first_tloc_locus_node;
   while (cur_tloc->p_next && cur_tloc->no < t_no)
      cur_tloc = cur_tloc->p_next;
   /*                                    * Adds params for the trait to line. */
   sprintf (p_lin + (n % 4) * 20, " T%d  ", t_no);

   val = *(cur_tloc->p_freq_node->p_freq);
   if (alle_frq_norm)
      val /= cur_tloc->p_freq_node->sum;
   sprintf (p_lin + (n % 4) * 20 + 5, "%5.3f", val);
   *(p_lin + (n % 4) * 20 + 5) = ' ';
   val = *(cur_tloc->p_freq_node->p_freq + 1);
   if (alle_frq_norm)
      val /= cur_tloc->p_freq_node->sum;
   sprintf (p_lin + (n % 4) * 20 + 10, "%5.3f", val);
   *(p_lin + (n % 4) * 20 + 10) = ' ';

   return;
}

/* Formats the genotype means for the trait at the specified tloc.            */

void fmt_trtl_geno_means (int t_no, int tl_no)
{
   GENO_MEANS_PTR p_m;
   char  prnt_lin[81] = "";
   static int prev_t_no = 0;

   double *means;
   double mini;
   double maxi;

   int   num_lines;
   int   precision;
   int   i1;
   int   i2;

   /* Find the geno_means_node for the specified trait and tloc combination.  */

   p_m = first_geno_means_node;
   while ((p_m && p_m->p_next && (p_m->trait_no < t_no))
          || (p_m->trait_no == t_no && p_m->tloc_no < tl_no))
      p_m = p_m->p_next;

   if (p_m && p_m->trait_no == t_no && p_m->tloc_no == tl_no)
   {
      means = p_m->means;
      if (means != NULL)
      {
         mini = means[0];               /* Determine print format for means.  */
         for (i2 = 1; i2 < p_m->num_means; ++i2)
            if (means[i2] < mini)
               mini = means[i2];

         maxi = means[0];
         for (i2 = 1; i2 < p_m->num_means; ++i2)
            if (means[i2] > maxi)
               maxi = means[i2];

         if (mini == 0.0 && maxi == 0.0)
            precision = 1;
         else
         {
            precision = decWid (mini, maxi, 4);
            if (precision > 7)
               precision = 7;
         }

         if (p_m->joint == TRUE && p_m->num_means == 9)
         {
            sprintf (prnt_lin, "                                  ");
            sprintf (prnt_lin + 34, "tloc%4d", p_m->joined_tlno);
            fprintf (stdout, "%s\n", prnt_lin); /* Output the line to stdout. */
            num_lines = 3;
         }
         else if (p_m->joint == FALSE && p_m->num_means == 3)
            num_lines = 1;
         else
            num_lines = 0;

         for (i1 = 0; i1 < num_lines; ++i1)
         {
            if (prev_t_no != t_no)
            {
               sprintf (prnt_lin, "  %4d", p_m->trait_no);
               prev_t_no = t_no;
            }
            else
               sprintf (prnt_lin, "      ");

            if (i1 == 0)
            {
               sprintf (prnt_lin + 6, "  %4d ", p_m->tloc_no);
               if (p_m->joint == TRUE)
                  sprintf (prnt_lin + 13, "  (1 1) ");
               else
                  sprintf (prnt_lin + 13, "        ");
            }
            else if (i1 == 1)
               sprintf (prnt_lin + 6, "         (1 2) ");
            else if (i1 == 2)
               sprintf (prnt_lin + 6, "         (2 2) ");

            sprintf (prnt_lin + 21, " %9.*f", precision, means[0]);
            sprintf (prnt_lin + 31, " %9.*f", precision, means[1]);
            sprintf (prnt_lin + 41, " %9.*f", precision, means[2]);

            fprintf (stdout, "%s\n", prnt_lin); /* Output the line to stdout. */

            means += 3;
         }
      }
   }

   return;
}

/**
 * formats the variances for the trait
 */
void fmt_trait_vari (char *p_lin, int t_no)
{
   VARI_PTR cur_vari;
   double addi, resi;
   int   precision;

   cur_vari = first_addi_node;          /* finds additive vari for this trait */
   while (cur_vari->p_next && cur_vari->no < t_no)
      cur_vari = cur_vari->p_next;

   if (cur_vari->no != t_no)
      return;

   addi = cur_vari->vari;

   cur_vari = first_resi_node;          /* finds residual vari for this trait */
   while (cur_vari->p_next && cur_vari->no < t_no)
      cur_vari = cur_vari->p_next;

   if (cur_vari->no != t_no)
      return;

   resi = cur_vari->vari;

   if (addi == 0.0 && resi == 0.0)      /* determines dec. places for vari's  */
      precision = 1;
   else
   {
      precision = decWid (addi, resi, 4);
      if (precision > 7)
         precision = 7;
   }
   sprintf (p_lin, " %5d    ", t_no);
   sprintf (p_lin + 10, " %9.*f", precision, addi);
   sprintf (p_lin + 20, " %9.*f", precision, resi);

   return;
}

void alloc_recomb (void)
{
   PRecombMap = malloc (NLoci * sizeof (RMAP));
   alloc_chk (WHERE, PRecombMap);

   return;
}

void mak_recomb (PRDIST_PTR p_ploc, int loc, int sx)
{
   RMAP_PTR p_recomb;
   PRDIST_PTR p_bck, p_nxt;
   int   i;
   double d, r_a, r_b, r_star;
   /*                                    * sets up recombination map for      */
   p_recomb = PRecombMap + loc;         /* markers & linked tlocs             */
   p_recomb->no = p_ploc->no;
   p_recomb->iA = p_recomb->iB = 0;
   p_recomb->probA[sx] = p_recomb->probB[sx] = 0.;

   if (p_ploc->locus_typ == MARK_LOC)   /*       **marker**                   */
   {
      p_bck = p_ploc;
      i = loc;
      /*                                 * looks for preceding marker        */
      while ((p_bck = p_bck->p_prev) && p_bck->locus_typ != MARK_LOC)
         --i;

      if (!p_bck || p_bck->locus_typ != MARK_LOC)
         p_recomb->recombTyp = FM;      /* 1st marker on chromosome           */

      else
      {
         p_recomb->recombTyp = OM;      /* not 1st marker on chromosome       */
         p_recomb->iA = i - 1;
         p_recomb->probA[sx] = dist2rf (p_ploc->dist);
      }
   }
   else if (!p_ploc->p_prev)            /*         **tloc**                   */
   {
      p_recomb->recombTyp = FT;         /* tloc is 1st locus on chromosome    */
      p_recomb->probB[sx] = dist2rf (p_ploc->dist);

      p_nxt = p_ploc;
      i = loc;
      /*                                 * finds following marker             */
      while ((p_nxt = p_nxt->p_next) && p_nxt->locus_typ != MARK_LOC)
         ++i;

      p_recomb->iB = i + 1;
   }
   else
      /*                                 * tloc is preceded by marker or tloc */
   {
      p_recomb->iA = loc - 1;

      p_nxt = p_ploc;
      i = loc;
      d = 0.;
      /*                                 * looks for following marker         */
      while ((p_nxt = p_nxt->p_next) && p_nxt->locus_typ != MARK_LOC)
      {
         ++i;
         d += p_nxt->loc_dist;
      }
      if (p_nxt && p_nxt->locus_typ == MARK_LOC)
      {                                 /* tloc has following marker          */
         p_recomb->recombTyp = IT;
         p_recomb->iB = i + 1;
         /*                              * probA: prob recomb if kA==kB       */
         /*                              * probB: prob recomb with B          */
         /*                              *        if kA!=kB                   */
         d += p_nxt->loc_dist;
         r_a = dist2rf (p_ploc->loc_dist);
         r_b = dist2rf (d);
         r_star = dist2rf (p_ploc->loc_dist + d);
         p_recomb->probA[sx] = r_a * r_b / (1 - r_star);
         if (r_star != 0.)
            p_recomb->probB[sx] = (1 - r_a) * r_b / r_star;
         else
            p_recomb->probB[sx] = 0.;
      }
      else
         /*                              * tloc has no following marker       */
      {
         p_recomb->recombTyp = ET;
         p_recomb->probA[sx] = dist2rf (p_ploc->loc_dist);
      }
   }
   return;
}

/**
 * sets up recomb map for unlinked tlocs
 */
void mak_no_recomb (int tl_no, int loc, int sx)
{
   RMAP_PTR p_recomb;

   p_recomb = PRecombMap + loc;
   p_recomb->no = tl_no;
   p_recomb->recombTyp = UT;
   p_recomb->iA = p_recomb->iB = 0;
   p_recomb->probA[sx] = p_recomb->probB[sx] = 0.;

   return;
}

/**
 * Allocates the trait parameters data structures.
 */
void alloc_traits (int n_traitPars)
{
   PTraitPars = malloc (n_traitPars * sizeof (TPARS));
   alloc_chk (WHERE, PTraitPars);

   return;
}

/**
 * allocates the tloc parameters data structures.
 */
void alloc_tloc_out (void)
{
   PTLocOut = malloc (NTLocs * sizeof (TLOUT));
   alloc_chk (WHERE, PTLocOut);

   return;
}

/**
 * Stores parameters for computing and writing the trait to the output
 * pedigree file:
 * the additive and residual variances,
 * where to find the genotype for the trait at the tloc,
 * the genotype means for the trait at this tloc,
 * format for the trait.
 */
void mak_trait (int t_no, int tl_no, int loc, int t_ndx)
{
   TPARS_PTR p_trait;
   GENO_MEANS_PTR p_m;
   VARI_PTR p_vari;
   double mini;
   double maxi;

   p_trait = PTraitPars + t_ndx;
   p_trait->no = t_no;
   p_trait->tlno = tl_no;
   p_trait->ixG = loc;
   p_trait->ixPh = t_ndx;

   p_vari = first_addi_node;            /* Next, store variances.             */
   while (p_vari->p_next && p_vari->no < t_no)
      p_vari = p_vari->p_next;
   p_trait->addi_var = p_vari->vari;
   p_trait->addi_term_found = sqrt (p_vari->vari);
   p_trait->addi_term = sqrt (p_vari->vari / 2.);

   p_vari = first_resi_node;
   while (p_vari->p_next && p_vari->no < t_no)
      p_vari = p_vari->p_next;
   p_trait->resi_var = p_vari->vari;
   p_trait->resi_term = sqrt (p_vari->vari);

   p_m = first_geno_means_node;         /* Next, store the genotype means.    */
   while (p_m->p_next && (p_m->trait_no < t_no
                          || (p_m->trait_no == t_no && p_m->tloc_no < tl_no)))
      p_m = p_m->p_next;

   if (!p_m->joint)
   {
      p_trait->means[0] = p_m->means[0];
      p_trait->means[1] = p_m->means[1];
      p_trait->means[2] = p_m->means[2];

      mini = p_trait->means[0];
      if (p_trait->means[1] < mini)
         mini = p_trait->means[1];
      if (p_trait->means[2] < mini)
         mini = p_trait->means[2];
      maxi = p_trait->means[0];
      if (p_trait->means[1] > maxi)
         maxi = p_trait->means[1];
      if (p_trait->means[2] > maxi)
         maxi = p_trait->means[2];

      p_trait->dec = decWid (mini, maxi, SIGFIG);
      p_trait->wid = fldWid (mini, maxi, p_trait->dec) + 1;
   }

   return;
}

/**
 * stores parameters for writing the tloc latent values to the
 * output pedigree file.
 */
void mak_tloc_out (int tl_no, int loc, int n)
{
   TLOUT_PTR p_tlo;

   p_tlo = PTLocOut + n;
   p_tlo->no = tl_no;
   p_tlo->ixG = loc;

   return;
}

void mak_freq (PRNTMAP_PTR p_pmap, PRDIST_PTR p_ploc, int no, int loc)
{
   /*                                * stores an array of allele frequencies */
   /*                                * for the locus                         */
   int   i;
   double *p_c, sum;

   LOCUS_PTR p_locus;

   if (!PAlleleFreq)
   {
      PAlleleFreq = malloc (NLoci * sizeof (ALLELE));
      alloc_chk (WHERE, PAlleleFreq);
   }
   if (p_ploc && p_ploc->locus_typ == MARK_LOC)
   {
      p_locus = first_mark_locus_node;
      while (((p_locus->p_next && (p_locus->chrom_no < p_pmap->chrom_no))
              || (p_locus->chrom_no == p_pmap->chrom_no
                  && p_locus->no < p_ploc->no)))
         p_locus = p_locus->p_next;
   }
   else
   {
      p_locus = first_tloc_locus_node;
      while (p_locus->p_next && p_locus->no < no)
         p_locus = p_locus->p_next;
   }

   (PAlleleFreq + loc)->nAlleles = p_locus->no_alleles;
   (PAlleleFreq + loc)->pCumFreq
      = malloc (p_locus->no_alleles * sizeof (double));
   p_c = (PAlleleFreq + loc)->pCumFreq;
   alloc_chk (WHERE, p_c);
   /*                                * cumulative frequencies are stored      */
   sum = p_locus->p_freq_node->sum;

   *(p_c) = *(p_locus->p_freq_node->p_freq);
   if (alle_frq_norm)
      *(p_c) /= sum;
   for (i = 1; i < p_locus->no_alleles; ++i)
   {
      if (alle_frq_norm)
         *(p_c + i) = *(p_c + i - 1) + *(p_locus->p_freq_node->p_freq + i)
            / sum;
      else
         *(p_c + i) = *(p_c + i - 1) + *(p_locus->p_freq_node->p_freq + i);
   }

   return;
}

/**
 * for some programs, sets allele frequencies always normalized.
 */
void set_alle_frq_deflts (void)
{
   if (GENEDROP || MARKERDROP)
      alle_frq_norm = TRUE;

   alle_frq_deflts = FALSE;             /* No longer uses deflt allele freqs. */

   return;
}

/**
 * Prints header for unlinked tloc parameter section.
 */

void prnt_unltloc_head (void)
{
   if (PrintLevel >= 2)
   {
      if (sim_unlnkd_tlocs->no_tloc_sel > 1)
      {
         fprintf (stdout, " Unlinked tlocs\n");
         fprintf (stdout, " ..............\n\n");
      }
      else
      {
         fprintf (stdout, " Unlinked tloc\n");
         fprintf (stdout, " .............\n\n");
      }
   }

   return;
}

/**
 * Prints header for inheritance indicator section.
 */

void prnt_inher_head (void)
{
   if (PrintLevel >= 2)
   {
      if (sim_traits->no_traits > 1)
      {
         fprintf (stdout, " Inheritance indicators\n");
         fprintf (stdout, " ......................\n\n");
      }
      else
      {
         fprintf (stdout, " Inheritance indicator\n");
         fprintf (stdout, " .....................\n\n");
      }
   }

   return;
}

/**
 * Prints header for trait parameter section.
 */

void prnt_trait_head (void)
{
   if (PrintLevel >= 2)
   {
      if (sim_traits->no_traits > 1)
      {
         fprintf (stdout, " Traits\n");
         fprintf (stdout, " ......\n\n");
      }
      else
      {
         fprintf (stdout, " Trait\n");
         fprintf (stdout, " .....\n\n");
      }
   }

   return;
}

/**
 * prints header for chromosome parameter section
 */

void prnt_chrom_head (SIM_PTR p_sim)
{
   if (PrintLevel >= 4)
   {
      if (p_sim->chrom_no)
      {
         fprintf (stdout, " Chromosome %d\n", p_sim->chrom_no);
         fprintf (stdout, " .............\n");
      }
      else
      {
         fprintf (stdout, " Chromosome map\n");
         fprintf (stdout, " ..............\n");
      }
   }

   return;
}

/**
 * for this chromosome and gender, puts markers into "print map"
 * maps not present default to sex-averaged map, then F map
 */
PRNTMAP_PTR mak_markprmap (SIM_PTR p_s, int sex)
{
   int   entry2, entry3;
   DIST_PTR p_d;
   PRNTMAP_PTR p_m;
   PRDIST_PTR p_l;

   if (!(p_m = add_prntmap (p_s, sex))) /* adds marker to map for gender      */
      return NULL;

   entry2 = FALSE;
   entry3 = FALSE;

   if (p_s->no_mark_map == 1)
   {
      p_l = add_only_mark_locus (p_m);
      sel_mark_locus (p_s, p_l, &entry3);
   }
   else
      while ((p_d = nxt_sim_mark_dist (&entry2, sex, p_s))) /*use gend deflts */
      {
         p_l = add_mark_locus (p_m, p_d);
         sel_mark_locus (p_s, p_l, &entry3);
      }

   return p_m;
}

/**
 * checks distances between markers not < specified minimum
 */
int chk_markmindist (PRNTMAP_PTR p_m)
{
   int   entry;
   double d;
   PRDIST_PTR p_l;

   entry = FALSE;
   d = rf2dist (RF_bound);
   while ((p_l = nxt_prntmap_locus (&entry, p_m)))
      if (!test_loc_mindist (p_l, d))
         return FALSE;

   return TRUE;
}

/**
 * for this chromosome and gender, puts tlocs into "print map"
 * maps not present default to sex-averaged map, then F map
 */
void mak_tlocprmap (SIM_PTR p_s, PRNTMAP_PTR p_m, int sex)
{
   int   entry;
   DIST_PTR p_d;
   PRDIST_PTR p_l;

   entry = FALSE;                       /* finds distances between adj loci   */

   while ((p_d = nxt_sim_tl_dist (&entry, sex, p_s))) /* use gendr deflt rules */
      if (p_d->selectd)
      {
         p_l = add_tloc_locus (p_m, p_d);
         sel_tloc_locus (p_l);
      }

   return;
}

/**
 * determines distances between adjacent loci, whether markers or tlocs
 */
void fix_prmap_dist (PRNTMAP_PTR p_m)
{
   int   entry = FALSE;
   PRDIST_PTR p_l;

   while ((p_l = nxt_prntmap_locus (&entry, p_m)))
      fix_locus_dist (p_m, p_l);

   return;
}

/**
 * checks that tlocs which are between markers
 * are actually between ADJACENT markers
 */
int chk_tlocprmap (PRNTMAP_PTR p_m)
{
   int   entry;
   PRDIST_PTR p_l;

   entry = FALSE;
   while ((p_l = nxt_prntmap_locus (&entry, p_m)))
      if (!test_tloc_dist (p_m, p_l))
         return FALSE;

   return TRUE;
}

/* The trim_blanks subroutine takes as input a string of characters and       */
/* inserts a NUL character after the final non-blank character in the string. */
/* In effect, this removes all trailing blank characters from the end of the  */
/* passed string.                                                             */

void trim_blanks (char *string)
{
   int   x1;                            /* Subscript to the final non-blank   */
   /*                                      character in the passed string.    */

   x1 = strlen (string);
   if (x1 > 0)
   {
      while (string[x1 - 1] == ' ')
         --x1;
      string[x1] = NUL;
   }

   return;
}
