/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 * proc_poly() has as its input the results of parsing the parameter
 * statements by proc_pars() for the polyEM programs.  It checks this
 * data for consistency and completeness and, if there are no errors,
 * builds data structures used by univar, unibig, bivar or multivar.
 *                                                        MAJ Jul 1999
 */

int proc_poly (void)
{
   alloc_poly_arrays ();                /* Allocate arrays used by proc_poly. */

   if (!test_select ())                 /* Check for which traits and fixed   */
      if (proc_err ())                  /* effects have been selected.        */
         return FALSE;

   if (n_err)
      return proc_end ();

   if (!test_startff ())                /* Check start values for fixed       */
      if (proc_err ())                  /* effects.                           */
         return FALSE;

   if (!test_start_var ())              /* Check start values for variances   */
      if (proc_err ())                  /* and covariances.                   */
         return FALSE;

   if (!test_eigen ())                  /* Check eigenvalue specifications.   */
      if (proc_err ())
         return FALSE;

   if (!test_peeling ())                /* Check peeling sequence             */
      if (proc_err ())                  /* specifications.                    */
         return FALSE;

   if (n_err)
      return proc_end ();

   return TRUE;
}

/* -------- test fncs:  test results of processing input statements --------- */

/* ----------------------- test functions for polyem ------------------------ */

/* Checks traits and fixed effects selected.  Then stores this information in */
/* the requisite arrays.                                                      */

int test_select (void)
{
   int   entry, t_no, effno, ndx, x1, x2, x3;
   SIMTRAIT_PTR p_st;
   TRAITEFF_PTR p_te;

   TraitTyp = QUANTITATIVE;
   alloc_sel ();

   if ((p_st = sim_traits))
   {
      NTraits = p_st->no_traits;
      for (x1 = 0; x1 < NTraits; ++x1)
      {
         t_no = p_st->p_trait[x1];

         /* For polyEM programs any selected trait number cannot be greater   */
         /* than the number of quantitative variables input from the pedigree */
         /* file.                                                             */
         if (t_no <= NReals)
         {
            /* For polyEM programs the trait number is still the same as the  */
            /* record position.  There is no input pedigree record option.    */
            if (x1 < psx->max_trait)
               TrNo[x1] = t_no;         /* TrNo[]: trait number.              */
            seltrff[t_no - 1][0] = TRUE; /* Include the mean.                 */
            SelReal[t_no - 1] = x1 + 1;
            selff[0] = TRUE;

/*       Update the arrays to reflect any fixed effects selected by the user. */
            entry = FALSE;              /* Find the trait effects array.      */
            while ((p_te = nxt_tr_eff (&entry)) && p_te->trait_num < t_no)
               ;
/*                                      If this is the current trait number:  */
            if (p_te && p_te->trait_num == t_no)
            {
               for (x2 = 0; x2 < p_te->num_eff; ++x2)
               {
                  effno = p_te->eff_nums[x2];
                  if (effno <= NInts)
/*                                   Set all variables for each fixed effect. */
                  {
                     seltrff[t_no - 1][effno] = TRUE;
                     SelInt[effno - 1] = TRUE;
                     selff[effno] = TRUE; /* selff[]: fixed effects to be     */
                  }                     /* modelled.                          */
                  else
                     err_typ = ERR_FFF;
               }
            }
         }
         else
         {
            err_typ = ERR_TRF;
            --NTraits;
         }
      }
   }

   if (!err_typ)
   {
      if (NTraits < psx->min_trait || NTraits > psx->max_trait)
      {
         err_typ = ERR_STN;
         err_txt = psx->pnam;
      }
   }

   if (!err_typ)
   {
      SelInt[0] = 4;                    /* (sex)                              */

      for (x3 = 1, ndx = 4; x3 < NInts; ++x3)
         if (SelInt[x3])
            SelInt[x3] = ++ndx;         /* (fixed effects)                    */

      NColData = ndx + 1;
      cum_selff[0] = selff[0];
      NFix = 0;

      for (x3 = 1; x3 <= NInts; ++x3)
      {
         cum_selff[x3] = cum_selff[x3 - 1] + selff[x3];
         if (selff[x3])
         {
            ++NFix;
            if (NFix <= FIXEFF)
               FENo[NFix] = x3;         /* FENo[]:  fixed effect number.      */
         }
      }

      if (NFix + 1 > psx->max_fef)
      {
         err_typ = ERR_SFN;
         err_txt = psx->pnam;
      }

   }

   if (!err_typ)
   {
      x2 = 0;
      for (x1 = 0; x1 < NReals; ++x1)
         if (SelReal[x1])
         {
            ndx = 0;
            for (x3 = 0; x3 <= NInts; ++x3)
               if (selff[x3])
               {
                  if (x2 < psx->max_trait && ndx < FIXEFF)
                     FfDex[x2][ndx] = seltrff[x1][x3];
                  ++ndx;                /* FfDex[][]: fixed effects selected  */
               }                        /* for traits.                        */
            ++x2;
         }
   }

   if (!err_typ)
   {
      if (selff[1])                     /* Checks gender as a fixed effect.   */
         FESex = TRUE;
      else
         FESex = FALSE;
   }

   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

int test_startff (void)
/* Checks for starting values for fixed effects, and stores this information in
 * Mufti[][][].
 */
{
   int   entry, t_no, effno, x1, x2, x3;
   TRAITEFF_PTR p_te;
   FIXEDEFF_PTR p_fe;

   entry = FALSE;
   while ((p_te = nxt_tr_eff (&entry)))
   {
      t_no = p_te->trait_num;
      /*                                If this trait has been selected:      */
      if (!((x1 = (((t_no <= NReals) ? SelReal[t_no - 1] : 0) - 1)) < 0))
      {

         if (p_te->start_tr_mean_inp)
         {
            Mufti[x1][0][1] = p_te->start_tr_mean;
            MuftiStart[x1][0] = Y;
            if (!err_typ)
               err_typ = ERR_STV;
         }

         if (p_te->set_trait_eff_inp)
         {
            p_fe = p_te->p_eff_list;

            while (p_fe)
            {
               effno = p_fe->eff_num;
               if (seltrff[t_no - 1][effno])
               {
                  x2 = cum_selff[effno - 1];
                  for (x3 = 0; x3 < p_fe->nstartff; ++x3)
                     Mufti[x1][x2][x3 + 1] = p_fe->startff[x3];
                  MuftiStart[x1][x2] = Y;
                  if (!err_typ)
                     err_typ = ERR_STV;
               }
               else
                  err_typ = ERR_SEL;

               p_fe = p_fe->p_next;
            }
         }

      }
   }

   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

int test_start_var (void)
/* Checks variances.
 */
{
   int   i, j, row, col, ngroups, group[MAXQT + 1];
   double x[MAXQT * (MAXQT + 1) / 2];

   /*                                    * variance checks:                  */
   if (naddvar == 0)                    /* checks for starting values        */
      err_typ = ERR_VA0;
   if (!err_typ)
      if (nresvar == 0)
         err_typ = ERR_VR0;

   if (!err_typ)                        /* checks no of starting values      */
      if (naddvar != NTraits)
         err_typ = ERR_VSA;
   if (!err_typ)
      if (nresvar != NTraits)
         err_typ = ERR_VSR;

   if (!err_typ)                        /* checks for non-zero starting      */
      for (i = 1; i <= NTraits; ++i)    /* values                            */
         if (SIGSQ (i, i) == 0.0)
            err_typ = ERR_VAN;
   if (!err_typ)
      for (i = 1; i <= NTraits; ++i)
         if (TAUSQ (i, i) == 0.0)
            err_typ = ERR_VRN;

   /*                                    * covariance checks:                */
   if (!err_typ)                        /* checks # starting values          */
      if (naddcov > 0 && naddcov != NTraits * (NTraits - 1) / 2)
         err_typ = ERR_CSA;
   if (!err_typ)
      if (nrescov > 0 && nrescov != NTraits * (NTraits - 1) / 2)
         err_typ = ERR_CSR;

   if (!err_typ)                        /* converts correlations to          */
      if (add_cor_flag)                 /* covariances                       */
      {
         for (row = 2; row <= NTraits; ++row)
            for (col = 1; col < row; ++col)
               if (SIGSQ (row, row) >= 0.0 && SIGSQ (col, col) >= 0.0)
                  SIGSQ (row, col) *= sqrt (SIGSQ (row, row))
                     * sqrt (SIGSQ (col, col));
               else
                  SIGSQ (row, col) = 0.0;
      }
   if (!err_typ)
      if (res_cor_flag)
      {
         for (row = 2; row <= NTraits; ++row)
            for (col = 1; col < row; ++col)
               if (TAUSQ (row, row) >= 0.0 && TAUSQ (col, col) >= 0.0)
                  TAUSQ (row, col) *= sqrt (TAUSQ (row, row))
                     * sqrt (TAUSQ (col, col));
               else
                  TAUSQ (row, col) = 0.0;
      }

   /*                                    * fit flags:                        */
   if (!err_typ)                        /* checks no. flags                  */
      if (nfitadd > 0 && nfitadd != NTraits * (NTraits - 1) / 2)
         err_typ = ERR_CFA;
   if (!err_typ)
      if (nfitres > 0 && nfitres != NTraits * (NTraits - 1) / 2)
         err_typ = ERR_CFR;

   ngroups = 0;
   if (!err_typ)                        /* checks independent groups for     */
      for (i = 1; i <= NTraits; ++i)    /* consistency                       */
      {
         group[i] = 0;
         for (j = 1; j < i; ++j)
            if (SIG_MODL (i, j) != 0)
               group[i] = group[j];
         if (group[i] != 0)
         {
            for (j = 1; j < i; ++j)
               if (group[i] == group[j])
               {
                  if (SIG_MODL (i, j) != 1)
                     err_typ = ERR_CAF;
               }
               else
               {
                  if (SIG_MODL (i, j) != 0)
                     err_typ = ERR_CAF;
               }
         }
         else
         {
            ngroups = ngroups + 1;
            group[i] = ngroups;
         }
      }

   if (!err_typ)
      for (i = 1; i <= NTraits; ++i)
      {
         group[i] = 0;
         for (j = 1; j < i; ++j)
            if (TAU_MODL (i, j) != 0)
               group[i] = group[j];
         if (group[i] != 0)
         {
            for (j = 1; j < i; ++j)
               if (group[i] == group[j])
               {
                  if (TAU_MODL (i, j) != 1)
                     err_typ = ERR_CRF;
               }
               else
               {
                  if (TAU_MODL (i, j) != 0)
                     err_typ = ERR_CRF;
               }
         }
         else
         {
            ngroups = ngroups + 1;
            group[i] = ngroups;
         }
      }

   if (!err_typ)                        /* checks for starting values of 0.  */
      for (i = 1; i <= NTraits; ++i)    /* for independent covarances        */
      {
         for (j = 1; j < i; ++j)
            if (SIG_MODL (i, j) == 0 && SIGSQ (i, j) != 0.0)
               err_typ = ERR_CAZ;
      }

   if (!err_typ)
      for (i = 1; i <= NTraits; ++i)
      {
         for (j = 1; j < i; ++j)
            if (TAU_MODL (i, j) == 0 && TAUSQ (i, j) != 0.0)
               err_typ = ERR_CRZ;
      }

   if (!err_typ)                        /* checks for positive definite      */
      /*                                 * matrices                          */
      if (pos_def_test (SigSq, x, NTraits))
         err_typ = ERR_PDA;

   if (!err_typ)
      if (pos_def_test (TauSq, x, NTraits))
         err_typ = ERR_PDR;

   if (psx->max_trait == 1)             /* for univar & unibig, moves        */
   {                                    /* covarances to SigAdd & SigErr     */
      SigAdd = SigSq[0];
      SigErr = TauSq[0];
   }

   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

int pos_def_test (double *a, double *b, int n)
/* code taken from chfact in multivar
 */
{
   int   i, j, k, kk;
   double zz, piv;

   kk = n * (n + 1) / 2;
   for (i = n - 1; i >= 0; i--)
   {
      zz = a[--kk];
      for (k = i + 1; k < n; k++)
         zz -= BB (b, k, i) * BB (b, k, i);
      if (zz <= 0.0)
         return TRUE;
      b[kk] = piv = sqrt (zz);
      for (j = i - 1; j >= 0; j--)
      {
         zz = a[--kk];
         for (k = i + 1; k < n; k++)
            zz -= BB (b, k, i) * BB (b, k, j);
         b[kk] = zz / piv;
      }
   }
   return FALSE;
}

int test_eigen (void)
{

   if (eigen_set && eigen_compute)
   {                                    /* both "compute" and "set"          */
      err_typ = ERR_ET1;                /* eigenvalues                       */
      ICh = 1;                          /* sets to compute                   */
   }
   else if (eigen_compute)
      ICh = 1;                          /* sets to compute eigenvalues       */

   else if (eigen_set)
      ICh = 0;                          /* sets to eigenvalues were set      */

   else
   {                                    /* sets to compute eigenvalues       */
      err_typ = ERR_ET2;                /* (default)                         */
      ICh = 1;
   }

   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

int test_peeling (void)
/* determines whether a peeling sequence file was named in command line
 * or parameter statements
 */
{

   pft = get_filent (PEEL);

   if (pft->command)                    /* tests for peeling sequence file   */
   {                                    /* named in command line             */
      if (pft->param)
         err_typ = ERR_PT1;
   }
   else if (pft->param)                 /* tests for peeling sequence file   */
      ;                                 /* named in param statement          */

   if (!err_typ)
      return TRUE;
   else
      return FALSE;
}

void alloc_poly_arrays (void)
/* Allocates memory for the arrays used by proc_poly to keep track of which   */
/* quantitative traits and corresponding fixed effects have been selected or  */
/* not: seltrff, selff and cum_selff.  The first 2 of these arrays need to be */
/* initialized to all zeroes, so calloc should be used instead of malloc.     */
{
   int   nreals_p1, nints_p1, x1;

   nreals_p1 = NReals + 1;
   nints_p1 = NInts + 1;

   seltrff = (int **) malloc (nreals_p1 * sizeof (int *));
   alloc_chk (WHERE, seltrff);

   seltrff[0] = (int *) calloc (nreals_p1 * nints_p1, sizeof (int));
   alloc_chk (WHERE, seltrff[0]);

   for (x1 = 1; x1 <= NReals; ++x1)
      seltrff[x1] = seltrff[x1 - 1] + nints_p1;

   selff = (int *) calloc (nints_p1, sizeof (int));
   alloc_chk (WHERE, selff);

   cum_selff = (int *) malloc (nints_p1 * sizeof (int));
   alloc_chk (WHERE, cum_selff);

   return;
}

void free_poly_arrays (void)
/* Frees the arrays used by proc_poly for keeping track of which quantitative */
/* traits and corresponding fixed effects have been selected.                 */
{

   if (seltrff)
   {
      if (seltrff[0])
         free (seltrff[0]);
      free (seltrff);
      seltrff = NULL;
   }

   if (selff)
   {
      free (selff);
      selff = NULL;
   }

   if (cum_selff)
   {
      free (cum_selff);
      cum_selff = NULL;
   }

   return;
}
