/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#include "lsample.h"
#include "msample.h"

/**
 * Get initial xx0, xx1 using locus-by-locus sampling or sequential imputation.
 * Then, from these, set the initial FGLs (founder genome labels).
 *
 * As for sampling(), the mm-flag indicates whether locus 0 (mm=0) should
 *    be set/sampled or not (mm=1 or mm=2): EAT  July 2005.
 */
void initInherit (int mm)
{
   int   i1, i2;

   prntStatusSMC ();

   set_seg_mode (0);

   peel_trace = 0;

   /* Set up initial maternal and inheritance segregation indicators.
    * Use locus-by-locus sampling or sequential imputation.
    *                                            xx0[][], xx1[][].            */

   if (SmplSetup == INDEP_LOC)
   {
      set_lsmpl (0, mm, NLoci - 1);
      prntStatusIMC ("locus-by-locus sampling");
   }

   else if (SmplSetup == SEQIMP)
   {
      if (TraitTyp != QUANTITATIVE)
         do_seq_imp (NSeqImp, mm, PedSize, NLoci, 1.0);
      else
         do_seq_imp (NSeqImp, mm, PedSize, NLoci, loci[0].ve);
      prntStatusIMC ("sequential imputation");
   }

   peel_trace = 0;
   if (peel_trace)
   {
      printf ("\n check initial inheritance assignments \n");
      fflush (stdout);
      for (i2 = 0; i2 < PedSize; ++i2)
         if (!(isNullNode (nghd[i2].pa)))
         {
            printf (" %s", nghd[i2].self);
            fflush (stdout);
            for (i1 = mm; i1 < NLoci; ++i1)
               printf ("  %d %d", xx1[i1][i2], xx0[i1][i2]);
            printf ("\n \n");
            fflush (stdout);
         }
   }

   /* Set the initial FGLs:                      gg0[][], gg1[][].            */
   for (i1 = mm; i1 < NLoci; ++i1)
      set_FGL (i1);

   if (peel_trace)
   {
      printf ("\n check initial fgl assignments \n");
      fflush (stdout);
      for (i2 = 0; i2 < PedSize; ++i2)
      {
         printf (" %s", nghd[i2].self);
         fflush (stdout);
         for (i1 = mm; i1 < NLoci; ++i1)
            printf ("  %d %d", gg1[i1][i2], gg0[i1][i2]);
         printf ("\n \n");
         fflush (stdout);
      }
   }

   peel_trace = 0;

   return;
}

/* Set up founder genome label peeling.                                       */

void setFGLpeel (int mm)
{
   static int need_to_call_alloc_gen_pen = TRUE;
   static int need_to_call_alloc_seg_pen = TRUE;

   int   i1;

/* Allocate necessary variables for gen_pen() and seg_pen().                  */

   for (i1 = mm; i1 < NLoci; ++i1)
   {
      if (loci[i1].locus_type == TRAIT_LOCUS)
      {
         if (need_to_call_alloc_gen_pen)
         {
            alloc_gen_pen (nFGL, loci[i1].n_all);
            need_to_call_alloc_gen_pen = FALSE;
         }
      }
      else if (loci[i1].locus_type == MARKER_LOCUS)
      {
         if (need_to_call_alloc_seg_pen)
         {
            alloc_seg_pen (nFGL, NLoci);
            need_to_call_alloc_seg_pen = FALSE;
         }
      }
   }

   return;
}

/* Set up prob[][] array for storing seg_pen results.                         */

double **initPenProb (int mm)
{
   double **prob;                       /* Matrix of resulting probabilities. */
   int   comp_no;                       /* Current component number.          */
   int   i1;
   int   err;

   /* Allocate prob[][] array for storing seg_pen results.                    */
   ald_nc_nl (&prob);

   for (i1 = mm; i1 < NLoci; ++i1)
      for (comp_no = 0; comp_no < nCompnt; ++comp_no)
         prob[comp_no][i1] = seg_pen2 (comp_no, loci + i1, nFGL, &err);

   if (DebugMain)
      dbgProb (prob);

   return prob;
}
