/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/***************************************************************************/
/*   set_FGL() does setting of gg0/gg1 from xx0/xx1 at locus j on all cmpts */
/*   Note gene_update() does updating from a given meiosis on down          */
/*   Note set_FGL_c() in TwoQTL sets by component, but in different format  */
/*   we need a reset_FGL() by component and by locus: added here 12/2008    */
/*                                                              EAT         */
/***************************************************************************/
#include "markers.h"

void set_FGL (int j)                    /* set founder genome labels 
                                         * at locus j, 
                                         * given the gene switches */
{
   int   i, cg;
   Ind  *dad, *mum;

   cg = 0;
   for (i = 0; i < pedsize; ++i)
   {
      dad = nghd[i].pa;
      if (xx0[j][i] == -1 && isNullNode (dad))
      {
         ++cg;
         gg0[j][i] = cg;
         ++cg;
         gg1[j][i] = cg;
      }
      else
      {
         mum = nghd[i].ma;
         if ((xx0[j][i] == -1) || (xx0[j][i] == 0))
            gg0[j][i] = gg0[j][mum->index];
         else
            gg0[j][i] = gg1[j][mum->index];
         if ((xx1[j][i] == -1) || (xx1[j][i] == 0))
            gg1[j][i] = gg0[j][dad->index];
         else
            gg1[j][i] = gg1[j][dad->index];
      }
   }
   return;
}

void reset_FGL (int j, int cc)          /* update founder genome labels 
                                         * at locus j,  on cmpnt cmp
                                         * given the gene switches */
/* if cc = 0, do on all components: do NOT reset founders          */
{
   int   i;
   Ind  *dad, *mum;
   Indlist *pind;

   if (!cc)                             /* reset all cmpnts   */
      for (i = 0; i < pedsize; ++i)
      {
         if (!isNullNode (nghd[i].pa))  /* i a non-founder */
         {
            dad = nghd[i].pa;
            mum = nghd[i].ma;
            if ((xx0[j][i] == -1) || (xx0[j][i] == 0))
               gg0[j][i] = gg0[j][mum->index];
            else
               gg0[j][i] = gg1[j][mum->index];
            if ((xx1[j][i] == -1) || (xx1[j][i] == 0))
               gg1[j][i] = gg0[j][dad->index];
            else
               gg1[j][i] = gg1[j][dad->index];
         }
      }
   else
   {
      pind = cmplist[cc].people->next;
      while (pind)
      {
         i = pind->current->index;
         if (!isNullNode (nghd[i].pa))  /* i a non-founder */
         {
            dad = nghd[i].pa;
            mum = nghd[i].ma;
            if ((xx0[j][i] == -1) || (xx0[j][i] == 0))
               gg0[j][i] = gg0[j][mum->index];
            else
               gg0[j][i] = gg1[j][mum->index];
            if ((xx1[j][i] == -1) || (xx1[j][i] == 0))
               gg1[j][i] = gg0[j][dad->index];
            else
               gg1[j][i] = gg1[j][dad->index];
         }
         pind = pind->next;
      }
   }
   return;
}
